@@ -23,6 +23,9 @@ extern launchtable launchlist[];
 extern int maxlaunch;
 extern archivetable archivelist[];
 extern int maxarchive;
+# ifndef	_NOBROWSE
+extern char **browsevar;
+# endif
 #endif
 extern char *macrolist[];
 extern int maxmacro;
@@ -40,6 +43,8 @@ extern int fd_restricted;
 #if	FD >= 2
 extern char *progpath;
 extern char *progname;
+extern char *promptstr;
+extern int lcmdline;
 #endif
 extern int internal_status;
 #ifdef	_NOORIGSHELL
@@ -52,10 +57,10 @@ extern int win_x;
 extern int win_y;
 extern int hideclock;
 #endif
+extern int inruncom;
 #ifndef	DECLERRLIST
 extern char *sys_errlist[];
 #endif
-extern int inruncom;
 
 static VOID NEAR builtinerror __P_((char *[], char *, int));
 #ifdef	_NOORIGSHELL
@@ -65,10 +70,17 @@ static VOID NEAR hitkey __P_((int));
 #endif
 static VOID NEAR fputsmeta __P_((char *, FILE *));
 #ifndef	_NOARCHIVE
+# if	FD >= 2
+static int NEAR getlaunchopt __P_((int, char *[], char *, launchtable *));
+# endif
 static int NEAR setlaunch __P_((int, char *[]));
 static int NEAR setarch __P_((int, char *[]));
 static int NEAR printlaunch __P_((int, char *[]));
 static int NEAR printarch __P_((int, char *[]));
+# ifndef	_NOBROWSE
+static char **NEAR readargv __P_((char **, char **));
+static int NEAR custbrowse __P_((int, char *[]));
+# endif
 #endif
 static int NEAR getcommand __P_((char *));
 static int NEAR setkeybind __P_((int, char *[]));
@@ -96,6 +108,8 @@ static int NEAR evalmacro __P_((int, char *[]));
 # if	!MSDOS && !defined (_NOKANJICONV)
 static int NEAR kconv __P_((int, char *[]));
 # endif
+static int NEAR getinputstr __P_((int, char *[]));
+static int NEAR getyesno __P_((int, char *[]));
 #endif	/* FD >= 2 */
 #ifdef	_NOORIGSHELL
 static int NEAR printenv __P_((int, char *[]));
@@ -120,22 +134,32 @@ static int NEAR loadsource __P_((int, char *[]));
 #define	FUNCNAME	1
 #endif
 
-#define	ER_FEWMANYARG	1
-#define	ER_OUTOFLIMIT	2
-#define	ER_NOENTRY	3
-#define	ER_SYNTAXERR	4
-#define	ER_EXIST	5
-#define	ER_INVALDEV	6
-#define	ER_EVENTNOFOUND	7
 static char *builtinerrstr[] = {
 	"",
+#define	ER_FEWMANYARG	1
 	"Too few or many arguments",
+#define	ER_OUTOFLIMIT	2
 	"Out of limits",
+#define	ER_NOENTRY	3
 	"No such entry",
+#define	ER_SYNTAXERR	4
 	"Syntax error",
+#define	ER_EXIST	5
 	"Entry already exists",
+#define	ER_INVALDEV	6
 	"Invalid device",
+#define	ER_EVENTNOFOUND	7
 	"Event not found",
+#define	ER_UNKNOWNOPT	8
+	"Unknown option",
+#define	ER_NOARGSPEC	9
+	"No argument is specified",
+#define	ER_NOTINSHELL	10
+	"Cannot execute in shell mode",
+#define	ER_NOTINRUNCOM	11
+	"Cannot execute in startup file",
+#define	ER_NOTRECURSE	12
+	"Cannot execute recursively",
 };
 #define	BUILTINERRSIZ	((int)(sizeof(builtinerrstr) / sizeof(char *)))
 
@@ -145,6 +169,9 @@ static builtintable builtinlist[] = {
 	{setarch,	BL_ARCH},
 	{printlaunch,	BL_PLAUNCH},
 	{printarch,	BL_PARCH},
+# ifndef	_NOBROWSE
+	{custbrowse,	BL_BROWSE},
+# endif
 #endif
 	{setkeybind,	BL_BIND},
 	{printbind,	BL_PBIND},
@@ -167,6 +194,8 @@ static builtintable builtinlist[] = {
 # if	!MSDOS && !defined (_NOKANJICONV)
 	{kconv,		BL_KCONV},
 # endif
+	{getinputstr,	BL_READLINE},
+	{getyesno,	BL_YESNO},
 #endif	/* FD >= 2 */
 #ifdef	_NOORIGSHELL
 # if	FD >= 2
@@ -244,7 +273,7 @@ int init;
 	if (n >= n_line) n = 1;
 	if (++n >= n_line) {
 		fflush(stdout);
-		ttyiomode();
+		ttyiomode(1);
 		win_x = 0;
 		win_y = n_line - 1;
 		hideclock = 1;
@@ -260,7 +289,7 @@ FILE *fp;
 {
 	char *cp;
 
-	if (!arg) fputs("\"\"", fp);
+	if (!*arg) fputs("\"\"", fp);
 	else {
 		cp = killmeta(arg);
 		kanjifputs(cp, fp);
@@ -269,139 +298,274 @@ FILE *fp;
 }
 
 #ifndef	_NOARCHIVE
+# if	FD >= 2
+static int NEAR getlaunchopt(n, argv, opts, list)
+int n;
+char *argv[], *opts;
+launchtable *list;
+{
+	char *cp;
+	int i, c;
+
+	if (argv[n][0] != '-') return(0);
+	c = argv[n][1];
+	if (!c || c == '-') return(++n);
+	else if (!strchr(opts, c)) {
+		builtinerror(argv, argv[n], ER_UNKNOWNOPT);
+		return(-1);
+	}
+
+	if (argv[n][2]) cp = &(argv[n][2]);
+	else if (argv[n + 1]) cp = argv[++n];
+	else {
+		builtinerror(argv, argv[n], ER_NOARGSPEC);
+		return(-1);
+	}
+
+	if (c == 'f') {
+		i = countvar(list -> format);
+		list -> format = (char **)realloc2(list -> format,
+			sizeof(char *) * (i + 2));
+		list -> format[i++] = strdup2(cp);
+		list -> format[i] = NULL;
+	}
+	else if (c == 'i') {
+		i = countvar(list -> lignore);
+		list -> lignore = (char **)realloc2(list -> lignore,
+			sizeof(char *) * (i + 2));
+		list -> lignore[i++] = strdup2(cp);
+		list -> lignore[i] = NULL;
+	}
+	else if (c == 'e') {
+		i = countvar(list -> lerror);
+		list -> lerror = (char **)realloc2(list -> lerror,
+			sizeof(char *) * (i + 2));
+		list -> lerror[i++] = strdup2(cp);
+		list -> lerror[i] = NULL;
+	}
+	else if (c == 't' || c == 'b') {
+		if ((i = atoi2(cp)) < 0) {
+			builtinerror(argv, cp, ER_SYNTAXERR);
+			return(-1);
+		}
+		if (c == 't') list -> topskip = i;
+		else list -> bottomskip = i;
+	}
+	else if (c == 'p') {
+		if (list -> ext) free (list -> ext);
+		list -> ext = strdup2(cp);
+	}
+	else {
+		if (!strcmp(cp, "loop"))
+			i = (c == 'd') ? LF_DIRLOOP : LF_FILELOOP;
+		else if (!strcmp(cp, "noprep"))
+			i = (c == 'd') ? LF_DIRNOPREP : LF_FILENOPREP;
+		else {
+			builtinerror(argv, cp, ER_SYNTAXERR);
+			return(-1);
+		}
+		list -> flags |= i;
+	}
+	n++;
+
+	return(n);
+}
+# endif
+
 static int NEAR setlaunch(argc, argv)
 int argc;
 char *argv[];
 {
-	char *ext;
-	long n;
-	int i, flags;
-# if	FD < 2
+# if	FD >= 2
+	int j, opt, err;
+# else
 	char *cp, *tmp;
-	int j, ch;
+	long l;
+	int ch;
 # endif
+	launchtable list;
+	int i, n;
 
 	if (argc <= 1) {
 		builtinerror(argv, NULL, ER_FEWMANYARG);
 		return(-1);
 	}
-# if	FD >= 2
-	if (argc >= 7) {
-		builtinerror(argv, NULL, ER_FEWMANYARG);
-		return(-1);
-	}
-# endif
 
-	flags = 0;
-	ext = getext(argv[1], &flags);
+	list.flags = 0;
+	list.ext = getext(argv[1], &(list.flags));
 
-	for (i = 0; i < maxlaunch; i++)
-		if (!extcmp(ext, flags,
-		launchlist[i].ext, launchlist[i].flags, 1)) break;
+	for (n = 0; n < maxlaunch; n++)
+		if (!extcmp(list.ext, list.flags,
+		launchlist[n].ext, launchlist[n].flags, 1)) break;
 
-	if (i >= MAXLAUNCHTABLE) {
+	if (n >= MAXLAUNCHTABLE) {
 		builtinerror(argv, NULL, ER_OUTOFLIMIT);
-		free(ext);
+		free(list.ext);
 		return(-1);
 	}
 
 	if (argc < 3) {
-		free(ext);
-		if (i >= maxlaunch) {
+		free(list.ext);
+		if (n >= maxlaunch) {
 			builtinerror(argv, argv[1], ER_NOENTRY);
 			return(-1);
 		}
 
-		free(launchlist[i].ext);
-		free(launchlist[i].comm);
+		free(launchlist[n].ext);
+		free(launchlist[n].comm);
 # if	FD >= 2
-		if (launchlist[i].format) free(launchlist[i].format);
+		freevar(launchlist[n].format);
+		freevar(launchlist[n].lignore);
+		freevar(launchlist[n].lerror);
 # endif
 		maxlaunch--;
-		for (; i < maxlaunch; i++)
-			memcpy((char *)&(launchlist[i]),
-				(char *)&(launchlist[i + 1]),
+		for (; n < maxlaunch; n++)
+			memcpy((char *)&(launchlist[n]),
+				(char *)&(launchlist[n + 1]),
 				sizeof(launchtable));
 		return(0);
 	}
 
 # if	FD >= 2
-	if (argc <= 4) launchlist[i].topskip = 0;
-	else if ((n = atoi2(argv[4])) >= 0) launchlist[i].topskip = n;
-	else {
-		builtinerror(argv, argv[4], ER_SYNTAXERR);
-		free(ext);
+	list.format = list.lignore = list.lerror = NULL;
+	list.comm = NULL;
+	list.topskip = list.bottomskip = 0;
+	opt = err = 0;
+	for (i = 2; i < argc; i++) {
+		j = getlaunchopt(i, argv, "fietb", &list);
+		if (!j) {
+			if (list.comm) break;
+			list.comm = strdup2(argv[i]);
+		}
+		else if (j > 0) {
+			opt++;
+			i = j - 1;
+		}
+		else {
+			err++;
+			break;
+		}
+	}
+
+	if (err);
+	else if (!(list.comm)) {
+		builtinerror(argv, NULL, ER_FEWMANYARG);
+		err++;
+	}
+	else if (i + 3 < argc) {
+		builtinerror(argv, argv[i + 3], ER_FEWMANYARG);
+		err++;
+	}
+	else if (i < argc && opt) {
+		builtinerror(argv, argv[i], ER_FEWMANYARG);
+		err++;
+	}
+	if (err) {
+		free(list.ext);
+		free(list.comm);
+		freevar(list.format);
+		freevar(list.lignore);
+		freevar(list.lerror);
 		return(-1);
 	}
-	if (argc <= 5) launchlist[i].bottomskip = 0;
-	else if ((n = atoi2(argv[5])) >= 0) launchlist[i].bottomskip = n;
+
+	if (i < argc) {
+		list.format = (char **)malloc2(sizeof(char *) * 2);
+		list.format[0] = strdup2(argv[i]);
+		list.format[1] = NULL;
+	}
+
+	if (i + 1 >= argc);
+	else if ((j = atoi2(argv[i + 1])) >= 0) list.topskip = j;
 	else {
-		builtinerror(argv, argv[5], ER_SYNTAXERR);
-		free(ext);
+		builtinerror(argv, argv[i + 1], ER_SYNTAXERR);
+		free(list.ext);
+		free(list.comm);
+		freevar(list.format);
+		freevar(list.lignore);
+		freevar(list.lerror);
 		return(-1);
 	}
+
+	if (i + 2 >= argc);
+	else if ((j = atoi2(argv[i + 2])) >= 0) list.bottomskip = j;
+	else {
+		builtinerror(argv, argv[i + 2], ER_SYNTAXERR);
+		free(list.ext);
+		free(list.comm);
+		freevar(list.format);
+		freevar(list.lignore);
+		freevar(list.lerror);
+		return(-1);
+	}
+
 # else	/* FD < 2 */
-	if (argc <= 3) launchlist[i].topskip = 255;
+	list.comm = strdup2(argv[2]);
+	if (argc <= 3) list.topskip = list.bottomskip = 255;
 	else {
 		cp = tmp = catvar(&(argv[3]), '\0');
-		if (!(cp = evalnumeric(cp, &n, 0)) || *cp != ',') {
+		if (!(cp = evalnumeric(cp, &l, 0)) || *cp != ',') {
 			builtinerror(argv, tmp, ER_SYNTAXERR);
-			free(ext);
+			free(list.ext);
+			free(list.comm);
 			free(tmp);
 			return(-1);
 		}
-		launchlist[i].topskip = n;
-		if (!(cp = evalnumeric(++cp, &n, 0))) {
+		list.topskip = l;
+		if (!(cp = evalnumeric(++cp, &l, 0))) {
 			builtinerror(argv, tmp, ER_SYNTAXERR);
-			free(ext);
+			free(list.ext);
+			free(list.comm);
 			free(tmp);
 			return(-1);
 		}
-		launchlist[i].bottomskip = n;
+		list.bottomskip = l;
 
 		ch = ':';
-		for (j = 0; j < MAXLAUNCHFIELD; j++) {
+		for (i = 0; i < MAXLAUNCHFIELD; i++) {
 			if (!cp || *cp != ch) {
 				builtinerror(argv, tmp, ER_SYNTAXERR);
-				free(ext);
+				free(list.ext);
+				free(list.comm);
 				free(tmp);
 				return(-1);
 			}
-			cp = getrange(++cp,
-				&(launchlist[i].field[j]),
-				&(launchlist[i].delim[j]),
-				&(launchlist[i].width[j]));
+			cp = getrange(++cp, &(list.field[i]),
+				&(list.delim[i]), &(list.width[i]));
 			ch = ',';
 		}
 
 		ch = ':';
-		for (j = 0; j < MAXLAUNCHSEP; j++) {
+		for (i = 0; i < MAXLAUNCHSEP; i++) {
 			if (*cp != ch) break;
-			if (!(cp = evalnumeric(++cp, &n, 0))) {
+			if (!(cp = evalnumeric(++cp, &l, 0))) {
 				builtinerror(argv, tmp, ER_SYNTAXERR);
-				free(ext);
+				free(list.ext);
+				free(list.comm);
 				free(tmp);
 				return(-1);
 			}
-			launchlist[i].sep[j] = (n > 0) ? n - 1 : 255;
+			list.sep[i] = (l > 0) ? l - 1 : 255;
 			ch = ',';
 		}
-		for (; j < MAXLAUNCHSEP; j++) launchlist[i].sep[j] = -1;
+		for (; i < MAXLAUNCHSEP; i++) list.sep[i] = -1;
 
-		if (!cp || *cp != ':') launchlist[i].lines = 1;
+		if (!cp || *cp != ':') list.lines = 1;
 		else {
-			if (!(cp = evalnumeric(++cp, &n, 0))) {
+			if (!(cp = evalnumeric(++cp, &l, 0))) {
 				builtinerror(argv, tmp, ER_SYNTAXERR);
-				free(ext);
+				free(list.ext);
+				free(list.comm);
 				free(tmp);
 				return(-1);
 			}
-			launchlist[i].lines = (n > 0) ? n : 1;
+			list.lines = (l > 0) ? l : 1;
 		}
 
 		if (*cp) {
 			builtinerror(argv, tmp, ER_SYNTAXERR);
-			free(ext);
+			free(list.ext);
+			free(list.comm);
 			free(tmp);
 			return(-1);
 		}
@@ -409,22 +573,17 @@ char *argv[];
 	}
 # endif	/* FD < 2 */
 
-	if (i >= maxlaunch) maxlaunch++;
+	if (n >= maxlaunch) maxlaunch++;
 	else {
-		free(launchlist[i].ext);
-		free(launchlist[i].comm);
+		free(launchlist[n].comm);
+		free(launchlist[n].ext);
 # if	FD >= 2
-		if (launchlist[i].format) free(launchlist[i].format);
+		freevar(launchlist[n].format);
+		freevar(launchlist[n].lignore);
+		freevar(launchlist[n].lerror);
 # endif
 	}
-	launchlist[i].ext = ext;
-	launchlist[i].comm = strdup2(argv[2]);
-# if	FD >= 2
-	launchlist[i].format = strdup2(argv[3]);
-# else
-	flags = 0;
-# endif
-	launchlist[i].flags = flags;
+	memcpy((char *)&(launchlist[n]), (char *)&list, sizeof(list));
 	return(0);
 }
 
@@ -433,14 +592,14 @@ int argc;
 char *argv[];
 {
 	char *ext;
-	int i, flags;
+	int i;
+	u_char flags;
 
 	if (argc <= 1 || argc >= 5) {
 		builtinerror(argv, NULL, ER_FEWMANYARG);
 		return(-1);
 	}
 
-	flags = 0;
 	ext = getext(argv[1], &flags);
 
 	for (i = 0; i < maxarchive; i++)
@@ -493,10 +652,8 @@ VOID printlaunchcomm(n, omit, fp)
 int n, omit;
 FILE *fp;
 {
-	char buf[MAXNUMCOLS + 1];
-# if	FD < 2
+	char buf[MAXLONGWIDTH + 1];
 	int i, ch;
-# endif
 
 	fputs(BL_LAUNCH, fp);
 	fputc(' ', fp);
@@ -508,16 +665,41 @@ FILE *fp;
 # if	FD >= 2
 	if (!launchlist[n].format) return;
 
-	fputc('\t', fp);
-	fputsmeta(launchlist[n].format, fp);
+	ch = 0;
+	for (i = 0; launchlist[n].format[i]; i++) {
+		if (i) {
+			fputs(" \\\n", fp);
+			ch++;
+		}
+		fputs("\t-f ", fp);
+		fputsmeta(launchlist[n].format[i], fp);
+	}
+	if (launchlist[n].lignore) for (i = 0; launchlist[n].lignore[i]; i++) {
+		fputs(" \\\n\t-i ", fp);
+		ch++;
+		fputsmeta(launchlist[n].lignore[i], fp);
+	}
+	if (launchlist[n].lerror) for (i = 0; launchlist[n].lerror[i]; i++) {
+		fputs(" \\\n\t-e ", fp);
+		ch++;
+		fputsmeta(launchlist[n].lerror[i], fp);
+	}
 	if (!launchlist[n].topskip && !launchlist[n].bottomskip) return;
-	fputc('\t', fp);
 
-	ascnumeric(buf, launchlist[n].topskip, 0, MAXNUMCOLS);
-	fputs(buf, fp);
-	fputc(' ', fp);
-	ascnumeric(buf, launchlist[n].bottomskip, 0, MAXNUMCOLS);
-	fputs(buf, fp);
+	if (launchlist[n].topskip) {
+		if (ch) fputs(" \\\n", fp);
+		fputs("\t-t ", fp);
+		ascnumeric(buf, launchlist[n].topskip, 0, MAXLONGWIDTH);
+		fputs(buf, fp);
+	}
+	if (launchlist[n].bottomskip) {
+		if (launchlist[n].topskip) fputc(' ', fp);
+		else if (ch) fputs(" \\\n\t", fp);
+		else fputc('\t', fp);
+		fputs("-b ", fp);
+		ascnumeric(buf, launchlist[n].bottomskip, 0, MAXLONGWIDTH);
+		fputs(buf, fp);
+	}
 # else	/* FD < 2 */
 	if (launchlist[n].topskip >= 255) return;
 	if (omit) {
@@ -525,10 +707,10 @@ FILE *fp;
 		return;
 	}
 	fputc('\t', fp);
-	ascnumeric(buf, launchlist[n].topskip, 0, MAXNUMCOLS);
+	ascnumeric(buf, launchlist[n].topskip, 0, MAXLONGWIDTH);
 	fputs(buf, fp);
 	fputc(',', fp);
-	ascnumeric(buf, launchlist[n].bottomskip, 0, MAXNUMCOLS);
+	ascnumeric(buf, launchlist[n].bottomskip, 0, MAXLONGWIDTH);
 	fputs(buf, fp);
 	ch = ':';
 	for (i = 0; i < MAXLAUNCHFIELD; i++) {
@@ -536,13 +718,13 @@ FILE *fp;
 		if (launchlist[n].field[i] >= 255) fputc('0', fp);
 		else {
 			ascnumeric(buf, launchlist[n].field[i] + 1,
-				0, MAXNUMCOLS);
+				0, MAXLONGWIDTH);
 			fputs(buf, fp);
 		}
 		if (launchlist[n].delim[i] >= 128) {
 			fputc('[', fp);
 			ascnumeric(buf, launchlist[n].delim[i] - 128,
-				0, MAXNUMCOLS);
+				0, MAXLONGWIDTH);
 			fputs(buf, fp);
 			fputc(']', fp);
 		}
@@ -555,7 +737,7 @@ FILE *fp;
 			fputc('-', fp);
 			if (launchlist[n].width[i] >= 128) {
 				ascnumeric(buf, launchlist[n].width[i] - 128,
-					0, MAXNUMCOLS);
+					0, MAXLONGWIDTH);
 				fputs(buf, fp);
 			}
 			else {
@@ -571,13 +753,13 @@ FILE *fp;
 		if (launchlist[n].sep[i] >= 255) break;
 		fputc(ch, fp);
 		ascnumeric(buf, launchlist[n].sep[i] + 1,
-			0, MAXNUMCOLS);
+			0, MAXLONGWIDTH);
 		fputs(buf, fp);
 		ch = ',';
 	}
 	if (launchlist[n].lines > 1) {
 		if (!i) fputc(':', fp);
-		ascnumeric(buf, launchlist[n].lines, 0, MAXNUMCOLS);
+		ascnumeric(buf, launchlist[n].lines, 0, MAXLONGWIDTH);
 		fputs(buf, fp);
 	}
 # endif	/* FD < 2 */
@@ -588,7 +770,8 @@ int argc;
 char *argv[];
 {
 	char *ext;
-	int i, flags;
+	int i;
+	u_char flags;
 
 	if (argc >= 3) {
 		builtinerror(argv, NULL, ER_FEWMANYARG);
@@ -601,7 +784,6 @@ char *argv[];
 		hitkey(0);
 	}
 	else {
-		flags = 0;
 		ext = getext(argv[1], &flags);
 		for (i = 0; i < maxlaunch; i++)
 			if (!extcmp(ext, flags,
@@ -639,7 +821,8 @@ int argc;
 char *argv[];
 {
 	char *ext;
-	int i, flags;
+	int i;
+	u_char flags;
 
 	if (argc >= 3) {
 		builtinerror(argv, NULL, ER_FEWMANYARG);
@@ -652,7 +835,6 @@ char *argv[];
 		hitkey(0);
 	}
 	else {
-		flags = 0;
 		ext = getext(argv[1], &flags);
 		for (i = 0; i < maxarchive; i++)
 			if (!extcmp(ext, flags,
@@ -668,6 +850,217 @@ char *argv[];
 	fflush(stdout);
 	return(0);
 }
+
+# ifndef	_NOBROWSE
+static char **NEAR readargv(sargv, dargv)
+char **sargv, **dargv;
+{
+	FILE *fp;
+	char *cp, *line, **argv;
+	ALLOC_T size;
+	int i, j, n, argc, dargc, meta, quote, quoted;
+
+	dargc = countvar(dargv);
+	for (n = 1; sargv[n]; n++) {
+		if (sargv[n][0] != '-' || sargv[n][1] != '@') {
+			dargv = (char **)realloc2(dargv,
+				(dargc + 2) * sizeof(char *));
+			dargv[dargc++] = strdup2(sargv[n]);
+			dargv[dargc] = NULL;
+			continue;
+		}
+		if (sargv[n][2]) cp = &(sargv[n][2]);
+		else if (sargv[n + 1]) cp = sargv[++n];
+		else {
+			builtinerror(sargv, sargv[n], ER_NOARGSPEC);
+			freevar(dargv);
+			return(NULL);
+		}
+		if (cp[0] == '-' && !cp[1]) fp = stdin;
+		else if (!(fp = Xfopen(cp, "r"))) {
+			builtinerror(sargv, cp, -1);
+			freevar(dargv);
+			return(NULL);
+		}
+
+		argc = 1;
+		argv = (char **)malloc2(2 * sizeof(char *));
+		argv[0] = strdup2(sargv[0]);
+		argv[1] = NULL;
+		j = meta = 0;
+		quote = quoted = '\0';
+		cp = c_realloc(NULL, 0, &size);
+		while ((line = fgets2(fp, 0))) {
+			if (!meta && !quote && *line == '#') {
+				free(line);
+				continue;
+			}
+			meta = 0;
+			for (i = 0; line[i]; i++) {
+				cp = c_realloc(cp, j + 2, &size);
+				if (line[i] == quote) {
+					quoted = quote;
+					quote = '\0';
+				}
+#  ifdef	CODEEUC
+				else if (isekana(line, i)) {
+					cp[j++] = line[i++];
+					cp[j++] = line[i];
+				}
+#  endif
+				else if (iskanji1(line, i)) {
+					cp[j++] = line[i++];
+					cp[j++] = line[i];
+				}
+				else if (quote) cp[j++] = line[i];
+				else if (line[i] == PMETA) {
+					if (!line[++i]) {
+						meta = 1;
+						break;
+					}
+					cp[j++] = line[i];
+				}
+				else if (line[i] == '\'' || line[i] == '"') {
+					quoted = '\0';
+					quote = line[i];
+				}
+				else if (!strchr(IFS_SET, line[i]))
+					cp[j++] = line[i];
+				else if (j || quoted) {
+					quoted = cp[j] = '\0';
+					argv = (char **)realloc2(argv,
+						(argc + 2) * sizeof(char *));
+					argv[argc++] = strdup2(cp);
+					argv[argc] = NULL;
+					j = 0;
+				}
+			}
+			if (meta);
+			else if (quote) cp[j++] = '\n';
+			else if (j || quoted) {
+				quoted = cp[j] = '\0';
+				argv = (char **)realloc2(argv,
+					(argc + 2) * sizeof(char *));
+				argv[argc++] = strdup2(cp);
+				argv[argc] = NULL;
+				j = 0;
+			}
+			free(line);
+		}
+		if (fp != stdin) Xfclose(fp);
+		else clearerr(fp);
+		if (j) {
+			cp[j] = '\0';
+			argv = (char **)realloc2(argv,
+				(argc + 2) * sizeof(char *));
+			argv[argc++] = strdup2(cp);
+			argv[argc] = NULL;
+		}
+		free(cp);
+		dargv = readargv(argv, dargv);
+		freevar(argv);
+		if (!dargv) return(NULL);
+		dargc = countvar(dargv);
+	}
+	return(dargv);
+}
+
+VOID freebrowse(list)
+launchtable *list;
+{
+	int i;
+
+	browselevel = 0;
+	freevar(browsevar);
+	browsevar = NULL;
+	if (argvar) {
+		for (i = 1; argvar[i]; i++) {
+			free(argvar[i]);
+			argvar[i] = NULL;
+		}
+	}
+	if (!list) return;
+	for (i = 0; list[i].comm; i++) {
+		if (list[i].ext) free(list[i].ext);
+		free(list[i].comm);
+		freevar(list[i].format);
+		freevar(list[i].lignore);
+		freevar(list[i].lerror);
+	}
+	free(list);
+}
+
+static int NEAR custbrowse(argc, argv)
+int argc;
+char *argv[];
+{
+	launchtable *list;
+	char **argv2;
+	int i, n, lvl;
+
+#  ifndef	_NOORIGSHELL
+	if (shellmode) {
+		builtinerror(argv, NULL, ER_NOTINSHELL);
+		return(-1);
+	}
+#  endif
+	if (inruncom) {
+		builtinerror(argv, NULL, ER_NOTINRUNCOM);
+		return(-1);
+	}
+	if (browselist) {
+		builtinerror(argv, NULL, ER_NOTRECURSE);
+		return(-1);
+	}
+	argv2 = (char **)malloc2(2 * sizeof(char *));
+	argv2[0] = strdup2(argv[0]);
+	argv2[1] = NULL;
+	if (!(argv2 = readargv(argv, argv2))) return(-1);
+
+	list = NULL;
+	lvl = 0;
+	n = 1;
+	while (argv2[n]) {
+		list = (launchtable *)realloc2(list,
+			(lvl + 2) * sizeof(launchtable));
+		list[lvl].topskip = list[lvl].bottomskip = 0;
+		list[lvl].comm = strdup2(argv2[n++]);
+		list[lvl].flags = 0;
+		list[lvl].ext = list[lvl + 1].comm = NULL;
+		list[lvl].format = list[lvl].lignore = list[lvl].lerror = NULL;
+		while (argv2[n]) {
+			i = getlaunchopt(n, argv2, "fietbpdn", &(list[lvl]));
+			if (!i) break;
+			else if (i > 0) n = i;
+			else {
+				freevar(argv2);
+				freebrowse(list);
+				return(-1);
+				break;
+			}
+		}
+		lvl++;
+	}
+	for (i = 0; i < lvl; i++)
+		if ((!i || i < lvl - 1) && !list[i].format) break;
+	if (!lvl || i < lvl) {
+		builtinerror(argv, NULL, ER_FEWMANYARG);
+		freevar(argv2);
+		freebrowse(list);
+		return(-1);
+	}
+	freevar(argv2);
+	if (list) {
+		freebrowse(browselist);
+		if (dolaunch(&(list[0]), 1) < 0) {
+			freebrowse(list);
+			return(-1);
+		}
+		browselist = list;
+	}
+	return(0);
+}
+# endif	/* !_NOBROWSE */
 #endif	/* !_NOARCHIVE */
 
 static int NEAR getcommand(cp)
@@ -918,6 +1311,9 @@ int head, sect, cyl;
 	int i, n, min, max;
 
 	min = -1;
+# ifdef	FAKEUNINIT
+	max = -1;	/* fake for -Wuninitialized */
+# endif
 	for (i = 0; fdtype[i].name; i++)
 	if (!strpathcmp(dev, fdtype[i].name)) {
 		if (min < 0) min = i;
@@ -960,12 +1356,15 @@ int head, sect, cyl;
 			if (drive > 'Z') break;
 			drvlist[i + 1] = drive;
 		}
-		if (j < n - 1) {
+		if (i < n - 1) {
 			free(sp);
 			free(drvlist);
 			return(-2);
 		}
 	}
+#  ifdef	FAKEUNINIT
+	else drvlist = NULL;		/* fake for -Wuninitialized */
+#  endif
 # endif	/* HDDMOUNT */
 
 	fdtype[max + n].name = NULL;
@@ -1320,7 +1719,7 @@ char *argv[];
 	}
 
 	hitkey(2);
-	ttyiomode();
+	ttyiomode(1);
 	i = 0;
 	for (;;) {
 		kanjiputs(GETKY_K);
@@ -1480,26 +1879,19 @@ char *argv[];
 			len += l2 - l1;
 			if (len > max) {
 				max = len;
+				j = cp - s;
 				s = realloc2(s, max + 1);
+				cp = s + j;
 			}
-			if (l1 > l2)
-				for (j = (cp - s) + l2; j <= len; j++)
-					s[j] = s[j - l2 + l1];
-			else if (l1 < l2)
-				for (j = len; j >= (cp - s) + l2; j--)
-					s[j] = s[j - l2 + l1];
+			memmove(&(cp[l2]), &(cp[l1]),
+				&(s[len]) - &(cp[l2]) + 1);
 			strncpy(cp, r, l2);
 		}
 		kanjifputs(s, stdout);
 		fputc('\n', stdout);
 		fflush(stdout);
 		entryhist(0, s, 0);
-		ttyiomode();
-		if ((ret = execmacro(s, NULL, 1, 1, 1)) < 0) {
-			internal_status = 1;
-			ret = 0;
-		}
-		stdiomode();
+		if ((ret = execmacro(s, NULL, 1, -1, 1)) < 0) ret = 0;
 		free(s);
 		return(ret);
 	}
@@ -1541,7 +1933,7 @@ char *argv[];
 	else {
 		int fd;
 
-		if ((fd = mktmpfile(path, argv[0])) < 0) {
+		if ((fd = mktmpfile(path)) < 0) {
 			builtinerror(argv, argv[0], -1);
 			free(tmp);
 			return(-1);
@@ -1549,7 +1941,7 @@ char *argv[];
 		if (!(fp = Xfdopen(fd, "w"))) {
 			builtinerror(argv, path, -1);
 			Xclose(fd);
-			rmtmpfile(strdup2(path));
+			rmtmpfile(path);
 			free(tmp);
 			return(-1);
 		}
@@ -1591,23 +1983,18 @@ char *argv[];
 	}
 
 	Xfclose(fp);
-	ttyiomode();
 	ret = execmacro(editor, path, 1, 0, 1);
-	stdiomode();
-	if (ret < 0) {
-		internal_status = 1;
-		ret = 0;
-	}
+	if (ret < 0) ret = 0;
 	if (ret) {
 		builtinerror(argv, editor, -1);
-		rmtmpfile(strdup2(path));
+		rmtmpfile(path);
 		free(tmp);
 		return(-1);
 	}
 
 	if (!(fp = Xfopen(path, "r"))) {
 		builtinerror(argv, path, -1);
-		rmtmpfile(strdup2(path));
+		rmtmpfile(path);
 		free(tmp);
 		return(-1);
 	}
@@ -1622,13 +2009,11 @@ char *argv[];
 		fputc('\n', stdout);
 		fflush(stdout);
 		entryhist(0, cp, 0);
-		ttyiomode();
-		ret = execmacro(cp, NULL, 1, 1, 1);
-		stdiomode();
+		ret = execmacro(cp, NULL, 1, -1, 1);
 		free(cp);
 	}
 	Xfclose(fp);
-	rmtmpfile(strdup2(path));
+	rmtmpfile(path);
 	return(ret);
 }
 #endif	/* !NOPOSIXUTIL */
@@ -1718,7 +2103,7 @@ u_long sum[4], x[16];
 	s[1] = 11;
 	s[2] = 16;
 	s[3] = 23;
-	t[0] = 0xfffa3942; 	/* floor(4294967296.0 * fabs(sin(32.0))) */
+	t[0] = 0xfffa3942;	/* floor(4294967296.0 * fabs(sin(32.0))) */
 	t[1] = 0x8771f681;
 	t[2] = 0x6d9d6122;
 	t[3] = 0xfde5380c;
@@ -1863,6 +2248,7 @@ char *argv[];
 			return(-1);
 		}
 		fputc('\n', stdout);
+		fflush(stdout);
 		return(0);
 	}
 
@@ -1876,6 +2262,7 @@ char *argv[];
 		fputc('\n', stdout);
 		hitkey(0);
 	}
+	fflush(stdout);
 	return(ret);
 }
 
@@ -1886,14 +2273,8 @@ char *argv[];
 	char *cp;
 	int ret;
 
-	if (argc < 2) return(0);
-	cp = catvar(&(argv[1]), ' ');
-	ttyiomode();
-	if ((ret = execmacro(cp, NULL, 1, 1, 0)) < 0) {
-		internal_status = 1;
-		ret = 0;
-	}
-	stdiomode();
+	if (argc <= 1 || !(cp = catvar(&(argv[1]), ' '))) return(0);
+	if ((ret = execmacro(cp, NULL, 1, -1, 0)) < 0) ret = 0;
 	free(cp);
 	return(ret);
 }
@@ -1943,6 +2324,7 @@ char *argv[];
 			fputs(argv[0], stderr);
 			fputs("[-i inputcode] [-o outputcode] [filename]\n",
 				stderr);
+			fflush(stderr);
 			return(-1);
 		}
 	}
@@ -1983,13 +2365,55 @@ char *argv[];
 		fputc('\n', fpout);
 		free(dest);
 	}
+	fflush(fpout);
 
 	discardunitable();
 	if (fpin != stdin) Xfclose(fpin);
+	else clearerr(fpin);
 	if (fpout != stdout) Xfclose(fpout);
 	return(0);
 }
 # endif	/* !MSDOS && !_NOKANJICONV */
+
+static int NEAR getinputstr(argc, argv)
+int argc;
+char *argv[];
+{
+	char *s, *duppromptstr;
+	int wastty;
+
+	if (!(wastty = isttyiomode)) ttyiomode(1);
+	duppromptstr = promptstr;
+	promptstr = (argc >= 2) ? argv[1] : "";
+	s = inputstr(NULL, 0, -1, NULL, -1);
+	promptstr = duppromptstr;
+	if (!wastty) stdiomode();
+	if (!s) return(-1);
+
+	kanjifputs(s, stdout);
+	fputc('\n', stdout);
+	fflush(stdout);
+	free(s);
+	return(0);
+}
+
+static int NEAR getyesno(argc, argv)
+int argc;
+char *argv[];
+{
+	char *s;
+	int ret, wastty;
+
+	if (argc < 2) s = "";
+	else s = argv[1];
+
+	if (!(wastty = isttyiomode)) ttyiomode(1);
+	lcmdline = -1;
+	ret = yesno(s);
+	if (!wastty) stdiomode();
+	return((ret) ? 0 : -1);
+}
+
 #endif	/* FD >= 2 */
 
 #ifdef	_NOORIGSHELL
@@ -1997,7 +2421,10 @@ static int NEAR printenv(argc, argv)
 int argc;
 char *argv[];
 {
-	int i, len;
+# if	FD < 2
+	int len;
+# endif
+	int i;
 
 # if	FD >= 2
 	if (argc >= 2) {
@@ -2399,7 +2826,7 @@ char *argv[];
 		builtinerror(argv, NULL, ER_FEWMANYARG);
 		return(-1);
 	}
-	ttyiomode();
+	ttyiomode(1);
 	ret = loadruncom(argv[1], 1);
 	stdiomode();
 	if (ret < 0) {
@@ -2447,6 +2874,7 @@ char *argv[];
 		fputs(": ", stderr);
 		kanjifputs(RESTR_K, stderr);
 		fputc('\n', stderr);
+		fflush(stderr);
 		return(RET_NOTICE);
 	}
 	if (argc > 2 || !filelist || maxfile <= 0) {
@@ -2454,9 +2882,10 @@ char *argv[];
 		fputs(": ", stderr);
 		kanjifputs(ILFNC_K, stderr);
 		fputc('\n', stderr);
+		fflush(stderr);
 		return(RET_NOTICE);
 	}
-	ttyiomode();
+	ttyiomode(0);
 	internal_status = (*funclist[n].func)(argv[1]);
 	locate(0, n_line - 1);
 	stdiomode();
@@ -2469,7 +2898,7 @@ char *command;
 int comline, ignorelist;
 {
 	char *cp, **argv;
-	int i, n, argc;
+	int i, n, nl, argc;
 
 	n = -1;
 	command = skipspace(command);
@@ -2484,9 +2913,9 @@ int comline, ignorelist;
 			tflush();
 		}
 		hitkey((comline) ? 1 : -1);
-		stdiomode();
+		nl = stdiomode();
 		n = execbuiltin(i, argc, argv);
-		ttyiomode();
+		ttyiomode(nl);
 		if (n == RET_SUCCESS) hitkey(3);
 		else if (comline) {
 			hideclock = 1;
@@ -2498,9 +2927,9 @@ int comline, ignorelist;
 			locate(0, n_line - 1);
 			tflush();
 		}
-		stdiomode();
+		nl = stdiomode();
 		n = execinternal(i, argc, argv);
-		ttyiomode();
+		ttyiomode(nl);
 		if (n == RET_SUCCESS);
 		else if (comline) {
 			hideclock = 1;
@@ -2513,10 +2942,10 @@ int comline, ignorelist;
 			locate(0, n_line - 1);
 			tflush();
 		}
-		stdiomode();
+		nl = stdiomode();
 		n = setuserfunc(i, argv);
 		n = (n < 0) ? RET_NOTICE : RET_SUCCESS;
-		ttyiomode();
+		ttyiomode(nl);
 		if (n == RET_SUCCESS);
 		else if (comline) {
 			hideclock = 1;
@@ -2585,7 +3014,9 @@ VOID freedefine(VOID_A)
 		free(launchlist[i].ext);
 		free(launchlist[i].comm);
 # if	FD >= 2
-		if (launchlist[i].format) free(launchlist[i].format);
+		freevar(launchlist[i].format);
+		freevar(launchlist[i].lignore);
+		freevar(launchlist[i].lerror);
 # endif
 	}
 	for (i = 0; i < maxarchive; i++) {
