@@ -25,7 +25,11 @@
 #include <time.h>
 #endif
 
-#if	!defined (FD) && MSDOS
+#if	MSDOS && defined (_NOUSELFN) && !defined (_NODOSDRIVE)
+#define	_NODOSDRIVE
+#endif
+
+#if	MSDOS && !defined (FD)
 #include <dos.h>
 #define	VOL_FAT32	"FAT32"
 # ifdef	DJGPP
@@ -59,13 +63,13 @@ typedef union REGS	__dpmi_regs;
 # define	PTR_SEG(ptr)		FP_SEG(ptr)
 # define	PTR_OFF(ptr, ofs)	FP_OFF(ptr)
 # endif	/* !DJGPP */
-#endif	/* !FD && MSDOS */
+#endif	/* MSDOS && !FD */
 
 #if	MSDOS
 #include <io.h>
 #include "unixemu.h"
 #define	DOSCOMOPT	'/'
-#define	C_EOF		CTRL('z')
+#define	C_EOF		K_CTRL('Z')
 #else	/* !MSDOS */
 #include <sys/time.h>
 #include <sys/file.h>
@@ -80,7 +84,7 @@ typedef union REGS	__dpmi_regs;
 # include <utime.h>
 # endif
 #define	DOSCOMOPT	'-'
-#define	C_EOF		CTRL('d')
+#define	C_EOF		K_CTRL('D')
 #endif	/* !MSDOS */
 
 #ifndef	FD
@@ -170,68 +174,38 @@ extern u_char _openfile[];
 extern char *sys_errlist[];
 #endif
 
+#include "pathname.h"
+#include "system.h"
+
+#if	defined (DOSCOMMAND) \
+&& (!defined (FD) || (FD >= 2 && !defined (_NOORIGSHELL)))
+
 #ifdef	FD
 #include "term.h"
 #else	/* !FD */
 # if	MSDOS
 # include <conio.h>
-# define	cc_intr		CTRL('c')
+# define	cc_intr		K_CTRL('C')
 # define	K_CR		'\r'
 # else	/* !MSDOS */
 extern int ttyio;
 static int cc_intr = -1;
 # define	K_CR		'\n'
-#  ifdef	USETERMIOS
-#  include <termios.h>
-#  include <sys/ioctl.h>	/* for Linux libc6 */
-typedef struct termios	termioctl_t;
-#  define	tioctl(d, r, a)	((r) \
-				? tcsetattr(d, (r) - 1, a) : tcgetattr(d, a))
-#  define	REQGETP		0
-#  define	REQSETP		(TCSAFLUSH + 1)
-#  else		/* !USETERMIOS */
-#   ifdef	USETERMIO
-#   include <termio.h>
-typedef struct termio	termioctl_t;
-#   define	tioctl		ioctl
-#   define	REQGETP		TCGETA
-#   define	REQSETP		TCSETAF
-#   else	/* !USETERMIO */
-#   include <sgtty.h>
-typedef struct sgttyb	termioctl_t;
-#   define	tioctl		ioctl
-#   define	REQGETP		TIOCGETP
-#   define	REQSETP		TIOCSETP
-#   endif	/* !USETERMIO */
-#  endif	/* !USETERMIOS */
 # endif	/* !MSDOS */
-# ifdef	CTRL
-# undef	CTRL
-# endif
-#define	CTRL(c)		((c) & 037)
+#define	K_CTRL(c)	((c) & 037)
 #define	K_BS		010
 #define	n_line		24
 #define	t_clear		"\033[;H\033[2J"
 #define	c_left		"\010"
 #endif	/* !FD */
 
-#include "pathname.h"
-#include "system.h"
-
-#if	!defined (_NOORIGSHELL) && defined (DOSCOMMAND) \
-&& (!defined (FD) || (FD >= 2))
-
 #ifdef	FD
-extern DIR *_Xopendir __P_((char *, int));
-#define	Xopendir(p)	_Xopendir(p, 0)
+extern DIR *Xopendir __P_((char *));
 extern int Xclosedir __P_((DIR *));
-extern struct dirent *_Xreaddir __P_((DIR *, int));
-#define	Xreaddir(d)	_Xreaddir(d, 0)
-extern char *_Xgetwd __P_((char *, int));
-#define	Xgetwd(p)	_Xgetwd(p, 0)
+extern struct dirent *Xreaddir __P_((DIR *));
+extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
-extern int _Xlstat __P_((char *, struct stat *, int, int));
-#define	Xlstat(p, s)	_Xlstat(p, s, 0, 0)
+extern int Xlstat __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
 extern int Xsymlink __P_((char *, char *));
 extern int Xreadlink __P_((char *, char *, int));
@@ -241,12 +215,9 @@ extern int Xutime __P_((char *, struct utimbuf *));
 # else
 extern int Xutimes __P_((char *, struct timeval []));
 # endif
-extern int _Xunlink __P_((char *, int));
-#define	Xunlink(p)	_Xunlink(p, 0)
-extern int _Xrename __P_((char *, char *, int));
-#define	Xrename(f, t)	_Xrename(f, t, 0)
-extern int _Xopen __P_((char *, int, int, int));
-#define	Xopen(p, f, m)	_Xopen(p, f, m, 0)
+extern int Xunlink __P_((char *));
+extern int Xrename __P_((char *, char *));
+extern int Xopen __P_((char *, int, int));
 # ifdef	_NODOSDRIVE
 # define	Xclose		close
 # define	Xread		read
@@ -256,10 +227,8 @@ extern int Xclose __P_((int));
 extern int Xread __P_((int, char *, int));
 extern int Xwrite __P_((int, char *, int));
 # endif	/* !_NODOSDRIVE */
-extern int _Xmkdir __P_((char *, int, int, int));
-#define	Xmkdir(p, m)	_Xmkdir(p, m, 0, 0)
-extern int _Xrmdir __P_((char *, int, int));
-#define	Xrmdir(p)	_Xrmdir(p, 0, 0)
+extern int Xmkdir __P_((char *, int));
+extern int Xrmdir __P_((char *));
 extern int chdir3 __P_((char *));
 extern int kanjifputs __P_((char *, FILE *));
 #else	/* !FD */
@@ -321,7 +290,7 @@ extern int chdir3 __P_((char *));
 
 struct filestat_t {
 	char *nam;
-#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char *d_alias;
 #else
 #define	d_alias		nam
@@ -365,6 +334,7 @@ struct filestat_t {
 
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
+extern char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 extern char *strdup2 __P_((char *));
 extern char *ascnumeric __P_((char *, long, int, int));
 
@@ -378,14 +348,14 @@ extern char *inputstr __P_((char *, int, int, char *, int));
 static int NEAR getinfofs __P_((char *, long *, long *, long *));
 # if	MSDOS
 static char *NEAR realpath2 __P_((char *, char *, int));
-# define	ttyiomode()
+# define	ttyiomode(n)
 # define	stdiomode()
 # define	getkey2(n)	getch();
 # else	/* !MSDOS */
 # define	realpath2(p, r, f) \
 				realpath(p, r)
 static VOID NEAR ttymode __P_((int));
-# define	ttyiomode()	(ttymode(1))
+# define	ttyiomode(n)	(ttymode(1))
 # define	stdiomode()	(ttymode(0))
 static int NEAR getkey2 __P_((int));
 # endif	/* !MSDOS */
@@ -441,10 +411,6 @@ static int NEAR doscopy __P_((char *, char *, struct stat *, int, int, int));
 #define	b_realloc(ptr, n, type) \
 			(((n) % BUFUNIT) ? ((type *)(ptr)) \
 			: (type *)realloc2(ptr, b_size(n, type)))
-#define	c_malloc(size)	(malloc2((size) = BUFUNIT))
-#define	c_realloc(ptr, n, size) \
-			(((n) + 1 < (size)) \
-			? (ptr) : realloc2(ptr, (size) *= 2))
 #define	dir_isdir(sp)	(((((sp) -> mod) & S_IFMT) == S_IFDIR) ? 1 : 0)
 
 
@@ -542,7 +508,7 @@ long *totalp, *freep, *bsizep;
 		reg.h.dl = toupper2(*path) - 'A' + 1;
 		int21call(&reg, &sreg);
 		if (reg.x.ax == 0xffff || !reg.x.ax || !reg.x.cx || !reg.x.dx
-		&& reg.x.dx < reg.x.bx) return(-1);
+		|| reg.x.dx < reg.x.bx) return(-1);
 
 		*totalp = (long)(reg.x.dx);
 		*freep = (long)(reg.x.bx);
@@ -600,7 +566,7 @@ int on;
 	if (cc_intr < 0) {
 		struct tchars cc;
 
-		if (tioctl(ttyio, TIOCGETC, &cc) < 0) cc_intr = CTRL('c');
+		if (tioctl(ttyio, TIOCGETC, &cc) < 0) cc_intr = K_CTRL('C');
 		else cc_intr = cc.t_intrc;
 	}
 	if (on) {
@@ -676,19 +642,20 @@ char *def;
 int h;
 {
 	char *cp;
-	int i, c, size;
+	ALLOC_T size;
+	int i, c;
 
 	fputs(prompt, stdout);
 	fflush(stdout);
 
-	cp = c_malloc(size);
+	cp = c_realloc(NULL, 0, &size);
 	for (i = 0; (c = inputkey()) != '\n'; i++) {
 		if (c < 0) {
 			free(cp);
 			return(NULL);
 		}
 		if (c != K_BS) {
-			cp = c_realloc(cp, i, size);
+			cp = c_realloc(cp, i, &size);
 			cp[i] = c;
 			if (!isctl(c)) fputc(c, stdout);
 			else {
@@ -779,7 +746,7 @@ static int NEAR inputkey(VOID_A)
 {
 	int c;
 
-	ttyiomode();
+	ttyiomode(1);
 	c = getkey2(0);
 	stdiomode();
 	if (c == EOF) c = -1;
@@ -1407,7 +1374,7 @@ long *sump, *bsump;
 
 		dirlist = b_realloc(dirlist, n, struct filestat_t);
 		dirlist[n].nam = strdup2(dp -> d_name);
-#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 		dirlist[n].d_alias = (dp -> d_alias[0])
 			? strdup2(dp -> d_alias) : dirlist[n].nam;
 #endif
@@ -1553,7 +1520,7 @@ long *sump, *bsump;
 
 	for (n = 0; n < max; n++) {
 		free(dirlist[n].nam);
-#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 		if (dirlist[n].d_alias != dirlist[n].nam)
 			free(dirlist[n].d_alias);
 #endif
@@ -1607,7 +1574,7 @@ char *argv[];
 			*(++file) = '\0';
 			dir = buf;
 			file = &(argv[n][++i]);
-			if (!(*file)) file = "*";
+			if (!*file) file = "*";
 		}
 	}
 
@@ -1744,7 +1711,7 @@ char *argv[];
 			fputs("All files in directory will be deleted!\n",
 				stdout);
 			fflush(stdout);
-			ttyiomode();
+			ttyiomode(1);
 			buf = inputstr("Are you sure (Y/N)?", 0, 0, NULL, -1);
 			stdiomode();
 			if (!buf) return(RET_SUCCESS);
@@ -1944,7 +1911,7 @@ char *file, *src;
 			errno = 0;
 			return(-1);
 		}
-		if ((fd = Xopen(file, O_RDONLY, 0644)) < 0) return(fd);
+		if ((fd = Xopen(file, O_RDONLY, 0666)) < 0) return(fd);
 		c = (isatty(fd)) ? 0 : 1;
 		close(fd);
 	}
@@ -1980,7 +1947,7 @@ char *file, *src;
 
 	flags = (copyflag & CF_VERIFY) ? O_RDWR : O_WRONLY;
 	flags |= (O_BINARY | O_CREAT | O_TRUNC);
-	return(Xopen(file, flags, 0644));
+	return(Xopen(file, flags, 0666));
 }
 
 static int NEAR textread(fd, buf, size, bin)
@@ -2037,7 +2004,7 @@ int sbin, dbin, dfd;
 {
 	off_t ofs;
 	char buf[BUFSIZ], buf2[BUFSIZ];
-	int i, n, fd1, fd2, tty, retry, tmperrno;
+	int i, n, fd1, fd2, tty, retry, duperrno;
 
 #if	!MSDOS
 	if (dfd < 0 && (stp -> st_mode & S_IFMT) == S_IFLNK) {
@@ -2072,14 +2039,17 @@ int sbin, dbin, dfd;
 	if (isatty(fd1) || isatty(fd2)) sbin = CF_TEXT;
 	tty = isatty(fd2);
 
+#ifdef	FAKEUNINIT
+	duperrno = errno;	/* fake for -Wuninitialized */
+#endif
 	for (retry = 0; retry < 10; retry++) {
 		for (;;) {
 			if ((i = textread(fd1, buf, BUFSIZ, sbin)) <= 0) {
-				tmperrno = errno;
+				duperrno = errno;
 				break;
 			}
 			if ((i = safewrite(fd2, buf, i)) < 0) {
-				tmperrno = errno;
+				duperrno = errno;
 				break;
 			}
 			if (i < BUFSIZ) break;
@@ -2091,16 +2061,16 @@ int sbin, dbin, dfd;
 
 		for (;;) {
 			if ((i = textread(fd1, buf, BUFSIZ, sbin)) <= 0) {
-				tmperrno = errno;
+				duperrno = errno;
 				break;
 			}
 			if ((n = textread(fd2, buf2, BUFSIZ, CF_BINARY)) < 0) {
-				tmperrno = errno;
+				duperrno = errno;
 				i = -1;
 				break;
 			}
 			if (n != i || memcmp(buf, buf2, i)) {
-				tmperrno = EINVAL;
+				duperrno = EINVAL;
 				i = -1;
 				break;
 			}
@@ -2116,7 +2086,7 @@ int sbin, dbin, dfd;
 
 	if (i < 0) {
 		Xunlink(dest);
-		errno = tmperrno;
+		errno = duperrno;
 		return(dfd < 0 ? -1 : -2);
 	}
 
@@ -2317,7 +2287,8 @@ char *argv[];
 {
 	char *cp;
 	u_char ch;
-	int i, n, fd, size;
+	ALLOC_T size;
+	int i, n, fd;
 #if	MSDOS && !defined (LSI_C)
 	int omode;
 #endif
@@ -2339,7 +2310,7 @@ char *argv[];
 	}
 #endif	/* MSDOS && !LSI_C */
 
-	cp = c_malloc(size);
+	cp = c_realloc(NULL, 0, &size);
 	i = 0;
 	for (;;) {
 		if ((n = textread(fd, &ch, sizeof(ch), CF_TEXT)) < 0) break;
@@ -2353,7 +2324,7 @@ char *argv[];
 			i = 0;
 		}
 		else {
-			cp = c_realloc(cp, i, size);
+			cp = c_realloc(cp, i, &size);
 			cp[i++] = ch;
 		}
 	}
@@ -2369,4 +2340,4 @@ char *argv[];
 	}
 	return(RET_SUCCESS);
 }
-#endif	/* !_NOORIGSHELL && DOSCOMMAND && (!FD || (FD >= 2)) */
+#endif	/* DOSCOMMAND && (!FD || (FD >= 2 && !_NOORIGSHELL)) */
