@@ -1,7 +1,7 @@
 /*
  *	FD (File & Directory maintenance tool)
  *
- *	by T.Shirai <shirai@nintendo.co.jp>
+ *	by T.Shirai <shirai@unixusers.net>
  */
 
 #include <ctype.h>
@@ -27,28 +27,12 @@ extern unsigned _stklen = 0x5800;
 # endif
 #endif	/* !MSDOS */
 
-#ifndef	_NOORIGSHELL
+#ifdef	_NOORIGSHELL
+#define	Xexit2		exit2
+#else
 #include "system.h"
 #endif
 
-#if	defined (SIGARGINT) || defined (NOVOID)
-#define	sigarg_t	int
-#else
-#define	sigarg_t	void
-#endif
-
-#ifdef	SIGFNCINT
-#define	sigfnc_t	int
-#else
-# ifdef	NOVOID
-# define	sigfnc_t
-# else
-# define	sigfnc_t	void
-# endif
-#endif
-
-#define	sigcst_t	sigarg_t (*)__P_((sigfnc_t))
-
 #if	!MSDOS && !defined (_NOORIGSHELL) && !defined (NOJOB)
 extern VOID killjob __P_((VOID_A));
 #endif
@@ -67,9 +51,13 @@ extern int maxarchive;
 extern char fullpath[];
 extern char **history[];
 extern char *helpindex[];
+extern int sizeinfo;
 extern int subwindow;
 extern int win_x;
 extern int win_y;
+#ifndef	_NOCUSTOMIZE
+extern int custno;
+#endif
 extern char *deftmpdir;
 #ifndef	_NODOSDRIVE
 extern char *unitblpath;
@@ -116,8 +104,11 @@ static int ignore_int __P_((VOID_A));
 #ifdef	SIGQUIT
 static int ignore_quit __P_((VOID_A));
 #endif
+#ifdef	SIGCONT
+static int ignore_cont __P_((VOID_A));
+#endif
 #ifdef	SIGHUP
-static int hangup __P_((VOID_A));
+static int hanguperror __P_((VOID_A));
 #endif
 #ifdef	SIGILL
 static int illerror __P_((VOID_A));
@@ -179,6 +170,9 @@ int adjtty = 0;
 #endif
 int showsecond = 0;
 int hideclock = 0;
+#ifdef	SIGALRM
+int noalrm = 1;
+#endif
 u_short today[3] = {0, 0, 0};
 int fd_restricted = 0;
 #ifndef	_NOCUSTOMIZE
@@ -201,6 +195,10 @@ int inruncom = 0;
 int fdmode = 0;
 
 static int timersec = 0;
+#ifdef	SIGWINCH
+static int winchok = 0;
+static int winched = 0;
+#endif
 
 
 #if	MSDOS && !defined (PROTECTED_MODE)
@@ -230,35 +228,50 @@ char *s;
 	endterm();
 	fputc('\007', stderr);
 	errno = duperrno;
-	perror(s);
+	if (errno) perror2(s);
+	else {
+		fputs(s, stderr);
+		fputc('\n', stderr);
+	}
 	inittty(1);
+	keyflush();
+	prepareexitfd();
 #ifndef	_NOORIGSHELL
 # if	!MSDOS && !defined (NOJOB)
 	killjob();
 # endif
 	prepareexit(0);
 #endif
+#ifdef	DEBUG
+# if	!MSDOS
+	freeterment();
+# endif
+	if (ttyout && ttyout != stderr) fclose(ttyout);
+	else if (ttyio >= 0) close(ttyio);
+	muntrace();
+#endif
+
 	exit(2);
 }
 
 static VOID NEAR signalexit(sig)
 int sig;
 {
-	signal(sig, SIG_IGN);
+	signal2(sig, SIG_IGN);
 	forcecleandir(deftmpdir, tmpfilename);
 #ifndef	_NODOSDRIVE
 	dosallclose();
 #endif
 	endterm();
 	inittty(1);
-	signal(sig, SIG_DFL);
+	signal2(sig, SIG_DFL);
 	kill(getpid(), sig);
 }
 
 #ifdef	SIGALRM
 static int ignore_alrm(VOID_A)
 {
-	signal(SIGALRM, (sigcst_t)ignore_alrm);
+	signal2(SIGALRM, (sigcst_t)ignore_alrm);
 	return(0);
 }
 #endif
@@ -266,7 +279,7 @@ static int ignore_alrm(VOID_A)
 #ifdef	SIGWINCH
 static int ignore_winch(VOID_A)
 {
-	signal(SIGWINCH, (sigcst_t)ignore_winch);
+	signal2(SIGWINCH, (sigcst_t)ignore_winch);
 	return(0);
 }
 #endif
@@ -274,7 +287,7 @@ static int ignore_winch(VOID_A)
 #ifdef	SIGINT
 static int ignore_int(VOID_A)
 {
-	signal(SIGINT, (sigcst_t)ignore_int);
+	signal2(SIGINT, (sigcst_t)ignore_int);
 	return(0);
 }
 #endif
@@ -282,13 +295,24 @@ static int ignore_int(VOID_A)
 #ifdef	SIGQUIT
 static int ignore_quit(VOID_A)
 {
-	signal(SIGQUIT, (sigcst_t)ignore_quit);
+	signal2(SIGQUIT, (sigcst_t)ignore_quit);
+	return(0);
+}
+#endif
+
+#ifdef	SIGCONT
+static int ignore_cont(VOID_A)
+{
+	signal2(SIGCONT, (sigcst_t)ignore_cont);
+# if	!MSDOS
+	suspended = 1;
+# endif
 	return(0);
 }
 #endif
 
 #ifdef	SIGHUP
-static int hangup(VOID_A)
+static int hanguperror(VOID_A)
 {
 	signalexit(SIGHUP);
 	return(0);
@@ -391,17 +415,63 @@ static int xsizerror(VOID_A)
 }
 #endif
 
+VOID checkscreen(xmax, ymax)
+int xmax, ymax;
+{
+	char *cp;
+	int i, row;
+
+	for (i = 0;; i++) {
+		if (!(cp = getwsize(xmax, ymax))) {
+#ifndef	_NOTREE
+			if (treepath) row = WFILEMINTREE;
+			else
+#endif
+#ifndef	_NOCUSTOMIZE
+			if (custno >= 0) row = WFILEMINCUSTOM;
+			else
+#endif
+			row = WFILEMIN;
+
+			if (FILEPERROW >= row) break;
+			cp = NOROW_K;
+		}
+		if (!i) {
+			putterm(t_clear);
+			locate(0, 0);
+			keyflush();
+		}
+
+		if (i & 1) cputs2(SCRSZ_K);
+		else {
+			putterm(t_standout);
+			cputs2(cp);
+			putterm(end_standout);
+		}
+		cputs2("\r\n");
+		putterm(t_bell);
+		tflush();
+		if (kbhit2(1000000L) && getkey2(0) == K_ESC) {
+			errno = 0;
+			error(INTR_K);
+		}
+	}
+}
+
 #ifdef	SIGWINCH
 static int wintr(VOID_A)
 {
 	int duperrno;
 
 	duperrno = errno;
-	signal(SIGWINCH, SIG_IGN);
-	getwsize(80, WHEADERMAX + WFOOTER + WFILEMIN);
-	rewritefile(1);
-	if (subwindow) ungetch2(CTRL('L'));
-	signal(SIGWINCH, (sigcst_t)wintr);
+	signal2(SIGWINCH, SIG_IGN);
+	if (!winchok) winched++;
+	else {
+		checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
+		rewritefile(1);
+		if (subwindow) ungetch2(K_CTRL('L'));
+	}
+	signal2(SIGWINCH, (sigcst_t)wintr);
 	errno = duperrno;
 	return(0);
 }
@@ -412,10 +482,10 @@ static int printtime(VOID_A)
 {
 	static time_t now;
 	struct tm *tm;
-	int duperrno;
+	int x, duperrno;
 
 	duperrno = errno;
-	signal(SIGALRM, SIG_IGN);
+	signal2(SIGALRM, SIG_IGN);
 	if (timersec) now++;
 	else {
 		now = time2();
@@ -437,7 +507,9 @@ static int printtime(VOID_A)
 		today[1] = tm -> tm_mon;
 		today[2] = tm -> tm_mday;
 		if (!hideclock) {
-			locate(n_column - 16 - ((showsecond) ? 3 : 0), LTITLE);
+			x = n_column - 15 - ((showsecond) ? 3 : 0);
+			if (!isleftshift()) x--;
+			locate(x, L_TITLE);
 			putterm(t_standout);
 			cprintf2("%02d-%02d-%02d %02d:%02d",
 				tm -> tm_year % 100,
@@ -450,7 +522,18 @@ static int printtime(VOID_A)
 		}
 	}
 	timersec--;
-	signal(SIGALRM, (sigcst_t)printtime);
+#ifdef	SIGWINCH
+	if (!winchok) winchok++;
+	else if (winched) {
+		signal2(SIGWINCH, SIG_IGN);
+		winched = 0;
+		checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
+		rewritefile(1);
+		if (subwindow) ungetch2(K_CTRL('L'));
+		signal2(SIGWINCH, (sigcst_t)wintr);
+	}
+#endif
+	signal2(SIGALRM, (sigcst_t)printtime);
 	errno = duperrno;
 	return(0);
 }
@@ -466,13 +549,15 @@ int set;
 	if (set == old);
 	else if (set) {
 #ifdef	SIGALRM
-		signal(SIGALRM, (sigcst_t)printtime);
+		signal2(SIGALRM, (sigcst_t)printtime);
+		noalrm = 0;
 #endif
 #ifdef	SIGTSTP
-		signal(SIGTSTP, SIG_IGN);
+		signal2(SIGTSTP, SIG_IGN);
 #endif
 #ifdef	SIGWINCH
-		signal(SIGWINCH, (sigcst_t)wintr);
+		winchok = winched = 0;
+		signal2(SIGWINCH, (sigcst_t)wintr);
 #endif
 #ifndef	_NODOSDRIVE
 		doswaitfunc = waitmes;
@@ -482,13 +567,14 @@ int set;
 	}
 	else {
 #ifdef	SIGALRM
-		signal(SIGALRM, (sigcst_t)ignore_alrm);
+		signal2(SIGALRM, (sigcst_t)ignore_alrm);
+		noalrm = 1;
 #endif
 #ifdef	SIGTSTP
-		signal(SIGTSTP, SIG_DFL);
+		signal2(SIGTSTP, SIG_DFL);
 #endif
 #ifdef	SIGWINCH
-		signal(SIGWINCH, (sigcst_t)ignore_winch);
+		signal2(SIGWINCH, (sigcst_t)ignore_winch);
 #endif
 #ifndef	_NODOSDRIVE
 		doswaitfunc = NULL;
@@ -503,11 +589,23 @@ int set;
 VOID title(VOID_A)
 {
 	char *cp, *eol;
-	int i;
+	int i, len;
 
-	locate(0, LTITLE);
+	locate(0, L_TITLE);
 	putterm(t_standout);
-	cputs2("  FD(File & Directory tool) Ver.");
+	len = 0;
+	if (!isleftshift()) {
+		putch2(' ');
+		len++;
+	}
+	cputs2(" FD");
+	len += 3;
+	if (!ishardomit()) {
+		cputs2("(File & Directory tool)");
+		len += 23;
+	}
+	cputs2(" Ver.");
+	len += 5;
 	cp = strchr(version, ' ');
 	while (*(++cp) == ' ');
 	if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
@@ -517,9 +615,9 @@ VOID title(VOID_A)
 		putch2('#');
 		i++;
 	}
-	cp = " (c)1995-2002 T.Shirai  ";
+	cp = (iswellomit()) ? "" : " (c)1995-2003 T.Shirai  ";
 	cputs2(cp);
-	i = n_column - 32 - strlen(cp) - i;
+	i = n_column - len - strlen(cp) - i;
 	while (i-- > 0) putch2(' ');
 	putterm(end_standout);
 	timersec = 0;
@@ -543,7 +641,7 @@ VOID saveorigenviron(VOID_A)
 		&origmaxarchive, maxarchive);
 # endif
 # if	!MSDOS && !defined (_NODOSDRIVE)
-	origfdtype = copydrive(NULL, fdtype);
+	origfdtype = copydosdrive(NULL, fdtype);
 # endif
 }
 #endif	/* !_NOCUSTOMIZE */
@@ -557,7 +655,7 @@ char *line;
 	char *cp;
 	int i;
 
-	if (!(*(cp = skipspace(command)))) i = 0;
+	if (!*(cp = skipspace(command))) i = 0;
 	else i = execpseudoshell(cp, 0, 1);
 	if (i) {
 		putterm(l_clear);
@@ -579,9 +677,9 @@ int exist;
 #ifdef	_NOORIGSHELL
 	FILE *fp;
 	char *cp, *tmp, *fold, *line;
-	int n, cont;
+	int cont;
 #endif
-	int er;
+	int n, er;
 
 #ifdef	_NOORIGSHELL
 # if	!MSDOS
@@ -653,9 +751,9 @@ int exist;
 	free(tmp);
 #else	/* !_NOORIGSHELL */
 	file = evalpath(strdup2(file), 1);
-	stdiomode();
+	n = stdiomode();
 	er = execruncom(file, 1);
-	ttyiomode();
+	ttyiomode(n);
 	free(file);
 #endif	/* !_NOORIGSHELL */
 
@@ -692,11 +790,8 @@ char *argv[], *envp[];
 		}
 	}
 
-#ifdef	_NOORIGSHELL
-	inittty(0);
-	getterment();
-#else
-	if (initshell(optc, optv, envp) < 0) exit2(RET_FAIL);
+#ifndef	_NOORIGSHELL
+	if (initshell(optc, optv, envp) < 0) Xexit2(RET_FAIL);
 #endif	/* !_NOORIGSHELL */
 	free(optv);
 	return(argc);
@@ -723,21 +818,15 @@ char *argv[];
 static VOID NEAR setexecname(argv)
 char *argv;
 {
-	char buf[MAXNAMLEN + 1];
+#if	MSDOS
+	char *cp;
+#endif
 
 	if ((progname = strrdelim(argv, 1))) progname++;
 	else progname = argv;
 #if	MSDOS
-	{
-		char *cp;
-
-		if ((cp = strchr(progname, '.'))) *cp = '\0';
-	}
-	sprintf(buf, "FD%ld", (long)getpid());
-#else
-	sprintf(buf, "%s%ld", progname, (long)getpid());
+	if ((cp = strchr(progname, '.'))) *cp = '\0';
 #endif
-	tmpfilename = strdup2(buf);
 }
 
 static VOID NEAR setexecpath(argv, envp)
@@ -774,7 +863,14 @@ char *argv, *envp[];
 
 static VOID NEAR prepareexitfd(VOID_A)
 {
-	if (_chdir2(origpath) < 0) error(origpath);
+	char cwd[MAXPATHLEN];
+
+	*cwd = '\0';
+	if (origpath && _chdir2(origpath) < 0) {
+		perror2(origpath);
+		if (!Xgetwd(cwd)) *cwd = '\0';
+		rawchdir(_SS_);
+	}
 	free(origpath);
 #ifndef	_NODOSDRIVE
 	dosallclose();
@@ -782,7 +878,7 @@ static VOID NEAR prepareexitfd(VOID_A)
 	free(progpath);
 
 #ifdef	DEBUG
-	free(tmpfilename);
+	if (tmpfilename) free(tmpfilename);
 # ifdef	_NOORIGSHELL
 	freevar(environ);
 	freevar(environ2);
@@ -808,7 +904,7 @@ static VOID NEAR prepareexitfd(VOID_A)
 #  endif
 #  if	!MSDOS && !defined (_NODOSDRIVE)
 	if (origfdtype) {
-		freedrive(origfdtype);
+		freedosdrive(origfdtype);
 		free(origfdtype);
 	}
 #  endif
@@ -820,19 +916,22 @@ static VOID NEAR prepareexitfd(VOID_A)
 # if	!MSDOS
 	freeidlist();
 # endif
+	chdir2(NULL);
 # if	!defined (_NOUSEHASH) && defined (_NOORIGSHELL)
 	searchhash(NULL, NULL, NULL);
 # endif
+# ifndef	_NOROCKRIDGE
+	detranspath(NULL, NULL);
+# endif
 #endif	/* DEBUG */
+	if (*cwd) rawchdir(cwd);
 }
 
 int main(argc, argv, envp)
 int argc;
 char *argv[], *envp[];
 {
-#ifndef	_NOORIGSHELL
 	char *cp;
-#endif
 	int i;
 
 #ifdef	DEBUG
@@ -849,49 +948,52 @@ char *argv[], *envp[];
 #endif
 
 #ifdef	SIGHUP
-	signal(SIGHUP, (sigcst_t)hangup);
+	signal2(SIGHUP, (sigcst_t)hanguperror);
 #endif
 #ifdef	SIGINT
-	signal(SIGINT, (sigcst_t)ignore_int);
+	signal2(SIGINT, (sigcst_t)ignore_int);
 #endif
 #ifdef	SIGQUIT
-	signal(SIGQUIT, (sigcst_t)ignore_quit);
+	signal2(SIGQUIT, (sigcst_t)ignore_quit);
+#endif
+#ifdef	SIGCONT
+	signal2(SIGCONT, (sigcst_t)ignore_cont);
 #endif
 #ifdef	SIGILL
-	signal(SIGILL, (sigcst_t)illerror);
+	signal2(SIGILL, (sigcst_t)illerror);
 #endif
 #ifdef	SIGTRAP
-	signal(SIGTRAP, (sigcst_t)traperror);
+	signal2(SIGTRAP, (sigcst_t)traperror);
 #endif
 #ifdef	SIGIOT
-	signal(SIGIOT, (sigcst_t)ioerror);
+	signal2(SIGIOT, (sigcst_t)ioerror);
 #endif
 #ifdef	SIGEMT
-	signal(SIGEMT, (sigcst_t)emuerror);
+	signal2(SIGEMT, (sigcst_t)emuerror);
 #endif
 #ifdef	SIGFPE
-	signal(SIGFPE, (sigcst_t)floaterror);
+	signal2(SIGFPE, (sigcst_t)floaterror);
 #endif
 #ifdef	SIGBUS
-	signal(SIGBUS, (sigcst_t)buserror);
+	signal2(SIGBUS, (sigcst_t)buserror);
 #endif
 #ifdef	SIGSEGV
-	signal(SIGSEGV, (sigcst_t)segerror);
+	signal2(SIGSEGV, (sigcst_t)segerror);
 #endif
 #ifdef	SIGSYS
-	signal(SIGSYS, (sigcst_t)syserror);
+	signal2(SIGSYS, (sigcst_t)syserror);
 #endif
 #ifdef	SIGPIPE
-	signal(SIGPIPE, (sigcst_t)pipeerror);
+	signal2(SIGPIPE, (sigcst_t)pipeerror);
 #endif
 #ifdef	SIGTERM
-	signal(SIGTERM, (sigcst_t)terminate);
+	signal2(SIGTERM, (sigcst_t)terminate);
 #endif
 #ifdef	SIGXCPU
-	signal(SIGXCPU, (sigcst_t)xcpuerror);
+	signal2(SIGXCPU, (sigcst_t)xcpuerror);
 #endif
 #ifdef	SIGXFSZ
-	signal(SIGXFSZ, (sigcst_t)xsizerror);
+	signal2(SIGXFSZ, (sigcst_t)xsizerror);
 #endif
 	sigvecset(0);
 
@@ -904,7 +1006,11 @@ char *argv[], *envp[];
 			strdup2(launchlist[maxlaunch].comm);
 # if	FD >= 2
 		launchlist[maxlaunch].format =
-			strdup2(launchlist[maxlaunch].format);
+			duplvar(launchlist[maxlaunch].format, -1);
+		launchlist[maxlaunch].lignore =
+			duplvar(launchlist[maxlaunch].lignore, -1);
+		launchlist[maxlaunch].lerror =
+			duplvar(launchlist[maxlaunch].lerror, -1);
 # endif
 	}
 	for (maxarchive = 0; maxarchive < MAXARCHIVETABLE; maxarchive++) {
@@ -935,6 +1041,12 @@ char *argv[], *envp[];
 	environ[i] = NULL;
 #endif
 
+#ifdef	_NOORIGSHELL
+	inittty(0);
+	getterment();
+#else
+	prepareterm();
+#endif
 	setexecpath(argv[0], envp);
 #ifndef	_NODOSDRIVE
 # ifdef	DATADIR
@@ -953,18 +1065,21 @@ char *argv[], *envp[];
 	if (*cp == 'r') cp++;
 # endif
 	if (!strpathcmp(cp, FDSHELL)) {
-		if (!Xgetwd(fullpath)) exit2(1);
+		if (!Xgetwd(fullpath)) Xexit2(1);
 		i = main_shell(argc, argv, envp);
 		prepareexitfd();
-		exit2(i);
+		Xexit2(i);
 	}
 #endif
 
 	fdmode = 1;
 	argc = initoption(argc, argv, envp);
-	ttyiomode();
+	ttyiomode(0);
 	initterm();
-	getwsize(80, WHEADERMAX + WFOOTER + WFILEMIN);
+	if ((cp = getwsize(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN))) {
+		errno = 0;
+		error(cp);
+	}
 #ifndef	_NOCUSTOMIZE
 	saveorigenviron();
 #endif
@@ -988,7 +1103,7 @@ char *argv[], *envp[];
 	if (adjtty) {
 		stdiomode();
 		inittty(0);
-		ttyiomode();
+		ttyiomode(0);
 	}
 #endif	/* !MSDOS */
 
@@ -996,6 +1111,9 @@ char *argv[], *envp[];
 	entryhist(1, origpath, 1);
 	putterms(t_clear);
 
+#ifdef	SIGWINCH
+	winchok = 1;
+#endif
 	main_fd(argv[i]);
 	sigvecset(0);
 	prepareexitfd();
@@ -1005,8 +1123,7 @@ char *argv[], *envp[];
 # if	!MSDOS && !defined (NOJOB)
 	killjob();
 # endif
-	prepareexit(0);
 #endif
-	exit2(0);
+	Xexit2(0);
 	return(0);
 }
