@@ -81,6 +81,7 @@ extern char *_mtrace_file;
 extern char *getenv2 __P_((char *));
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
+extern char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 extern char *strdup2 __P_((char *));
 extern char *strcpy2 __P_((char *, char *));
 extern char *strncpy2 __P_((char *, char *, int));
@@ -88,24 +89,23 @@ extern char *ascnumeric __P_((char *, long, int, int));
 # if	MSDOS || !defined (_NODOSDRIVE)
 extern int _dospath __P_((char *));
 #endif
-extern DIR *_Xopendir __P_((char *, int));
-#define	Xopendir(p)	_Xopendir(p, 0)
+extern DIR *Xopendir __P_((char *));
 extern int Xclosedir __P_((DIR *));
-extern struct dirent *_Xreaddir __P_((DIR *, int));
-#define	Xreaddir(d)	_Xreaddir(d, 0)
-extern char *_Xgetwd __P_((char *, int));
-#define	Xgetwd(p)	_Xgetwd(p, 0)
+extern struct dirent *Xreaddir __P_((DIR *));
+extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
 # if	MSDOS && !defined (_NOUSELFN)
 extern char *shortname __P_((char *, char *));
 # endif
+extern VOID demacroarg __P_((char **));
 extern char *progpath;
 #else	/* !FD */
 #define	getenv2		(char *)getenv
 static VOID NEAR allocerror __P_((VOID_A));
 char *malloc2 __P_((ALLOC_T));
 char *realloc2 __P_((VOID_P, ALLOC_T));
+char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 char *strdup2 __P_((char *));
 char *strcpy2 __P_((char *, char *));
 char *strncpy2 __P_((char *, char *, int));
@@ -135,6 +135,7 @@ char *Xgetwd __P_((char *));
 
 static char *NEAR getvar __P_((char *, int));
 static int NEAR setvar __P_((char *, char *, int));
+static int NEAR ismeta __P_((char *s, int, int, int));
 #ifdef	_NOORIGGLOB
 static char *NEAR cnvregexp __P_((char *, int));
 #else
@@ -220,10 +221,6 @@ static char *NEAR replacebackquote __P_((char *, int *, char *, int));
 #define	b_realloc(ptr, n, type) \
 			(((n) % BUFUNIT) ? ((type *)(ptr)) \
 			: (type *)realloc2(ptr, b_size(n, type)))
-#define	c_malloc(size)	(malloc2((size) = BUFUNIT))
-#define	c_realloc(ptr, n, size) \
-			(((n) + 1 < (size)) \
-			? (ptr) : realloc2(ptr, (size) *= 2))
 #define	getconstvar(s)	(getvar(s, sizeof(s) - 1))
 
 char **argvar = NULL;
@@ -355,8 +352,10 @@ u_char kctypetable[256] = {
 #endif	/* !LSI_C */
 
 static int skipdotfile = 0;
+#if	!defined (USEREGCMP) && !defined (USEREGCOMP) && !defined (USERE_COMP)
 static char *wildsymbol1 = "?";
 static char *wildsymbol2 = "*";
+#endif
 #if	!MSDOS
 static uidtable *uidlist = NULL;
 static int maxuid = 0;
@@ -378,7 +377,12 @@ ALLOC_T size;
 {
 	char *tmp;
 
-	if (!size || !(tmp = (char *)malloc(size))) allocerror();
+	if (!size || !(tmp = (char *)malloc(size))) {
+		allocerror();
+#ifdef	FAKEUNINIT
+		tmp = NULL;	/* fake for -Wuninitialized */
+#endif
+	}
 	return(tmp);
 }
 
@@ -390,10 +394,27 @@ ALLOC_T size;
 
 	if (!size
 	|| !(tmp = (ptr) ? (char *)realloc(ptr, size) : (char *)malloc(size)))
+	{
 		allocerror();
+#ifdef	FAKEUNINIT
+		tmp = NULL;	/* fake for -Wuninitialized */
+#endif
+	}
 	return(tmp);
 }
 
+char *c_realloc(ptr, n, sizep)
+char *ptr;
+ALLOC_T n, *sizep;
+{
+	if (!ptr) {
+		*sizep = BUFUNIT;
+		return(malloc2(*sizep));
+	}
+	while (n + 1 >= *sizep) *sizep *= 2;
+	return(realloc2(ptr, *sizep));
+}
+
 char *strdup2(s)
 char *s;
 {
@@ -423,7 +444,7 @@ int n;
 
 	for (i = 0; i < n && s2[i]; i++) s1[i] = s2[i];
 	s1[i] = '\0';
-	return(s1);
+	return(&(s1[i]));
 }
 
 /*
@@ -714,18 +735,13 @@ char *buf, *s1, *s2;
 	return(cp);
 }
 
-#if	!MSDOS
 int strcasecmp2(s1, s2)
 char *s1, *s2;
 {
-	int c1, c2;
-
 	for (;;) {
-		c1 = toupper2(*s1);
-		c2 = toupper2(*s2);
-		if (c1 != c2) return(c1 - c2);
+		if (toupper2(*s1) != toupper2(*s2)) return(*s1 - *s2);
 # ifndef	CODEEUC
-		if (issjis1(c1)) {
+		if (issjis1(*s1)) {
 			s1++;
 			s2++;
 			if (*s1 != *s2) return(*s1 - *s2);
@@ -737,13 +753,10 @@ char *s1, *s2;
 	}
 	return(0);
 }
-#endif	/* !MSDOS */
 
 int strpathcmp2(s1, s2)
 char *s1, *s2;
 {
-	int c1, c2;
-
 #if	!MSDOS
 	if (!pathignorecase) for (;;) {
 		if (*s1 != *s2) return(*s1 - *s2);
@@ -754,11 +767,9 @@ char *s1, *s2;
 	else
 #endif
 	for (;;) {
-		c1 = toupper2(*s1);
-		c2 = toupper2(*s2);
-		if (c1 != c2) return(c1 - c2);
+		if (toupper2(*s1) != toupper2(*s2)) return(*s1 - *s2);
 #ifndef	CODEEUC
-		if (issjis1(c1)) {
+		if (issjis1(*s1)) {
 			s1++;
 			s2++;
 			if (*s1 != *s2) return(*s1 - *s2);
@@ -775,8 +786,6 @@ int strnpathcmp2(s1, s2, n)
 char *s1, *s2;
 int n;
 {
-	int c1, c2;
-
 #if	!MSDOS
 	if (!pathignorecase) while (n-- > 0) {
 		if (*s1 != *s2) return(*s1 - *s2);
@@ -787,12 +796,10 @@ int n;
 	else
 #endif
 	while (n-- > 0) {
-		c1 = toupper2(*s1);
-		c2 = toupper2(*s2);
-		if (c1 != c2) return(c1 - c2);
+		if (toupper2(*s1) != toupper2(*s2)) return(*s1 - *s2);
 #ifndef	CODEEUC
-		if (issjis1(c1)) {
-			if (--n <= 0) break;
+		if (issjis1(*s1)) {
+			if (n-- <= 0) break;
 			s1++;
 			s2++;
 			if (*s1 != *s2) return(*s1 - *s2);
@@ -812,7 +819,8 @@ int len;
 	char *cp;
 
 	cp = malloc2(len + 1);
-	return(strncpy2(cp, s, len));
+	strncpy2(cp, s, len);
+	return(cp);
 }
 
 static char *NEAR getvar(ident, len)
@@ -870,28 +878,43 @@ char *s;
 	return(0);
 }
 
-int ismeta(s, ptr, quote)
+/*ARGSUSED*/
+char *isrootdir(s)
 char *s;
-int ptr, quote;
 {
-	if (s[ptr] != PMETA || !s[ptr + 1]) return(0);
-	if (quote) {
-		if (s[ptr + 1] != quote || s[ptr + 2]) return(1);
-		return(0);
-	}
-	return(1);
+#if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+	if (_dospath(s)) s += 2;
+#endif
+	if (*s != _SC_) return(NULL);
+	return(s);
 }
 
-int isnmeta(s, ptr, quote, len)
+static int NEAR ismeta(s, ptr, quote, len)
 char *s;
 int ptr, quote, len;
 {
-	if (s[ptr] != PMETA || ptr + 1 >= len) return(0);
-	if (quote) {
-		if (s[ptr + 1] != quote || ptr + 2 < len) return(1);
-		return(0);
+#ifdef	FAKEMETA
+	return(0);
+#else	/* !FAKEMETA */
+	if (s[ptr] != PMETA) return(0);
+	if (len >= 0) {
+		if (ptr + 1 >= len) return(0);
+		if (quote && s[ptr + 1] == quote && ptr + 2 >= len) return(0);
+	}
+	else {
+		if (!s[ptr + 1]) return(0);
+		if (quote) {
+			if (s[ptr + 1] == quote && !s[ptr + 2]) return(0);
+# if	MSDOS
+			if (!strchr(DQ_METACHAR, s[ptr + 1])) return(0);
+# endif
+		}
+# if	MSDOS
+		else if (!strchr(METACHAR, s[ptr + 1])) return(0);
+# endif
 	}
 	return(1);
+#endif	/* !FAKEMETA */
 }
 
 #ifdef	_NOORIGGLOB
@@ -900,25 +923,26 @@ char *s;
 int len;
 {
 	char *re;
-	int i, j, paren, size;
+	ALLOC_T size;
+	int i, j, paren;
 
 	if (!*s) {
 		s = "*";
 		len = 1;
 	}
 	if (len < 0) len = strlen(s);
-	re = c_malloc(size);
+	re = c_realloc(NULL, 0, &size);
 	paren = j = 0;
 	re[j++] = '^';
 	for (i = 0; i < len; i++) {
-		re = c_realloc(re, j + 2, size);
+		re = c_realloc(re, j + 2, &size);
 
 		if (paren) {
 			if (s[i] == ']') paren = 0;
 			re[j++] = s[i];
 			continue;
 		}
-		else if (isnmeta(s, i, '\0', len)) {
+		else if (ismeta(s, i, '\0', len)) {
 			re[j++] = s[i++];
 			re[j++] = s[i];
 		}
@@ -1071,7 +1095,8 @@ int len;
 {
 	reg_t *re;
 	char *cp, *paren;
-	int i, j, n, plen, size, metachar, quote;
+	ALLOC_T size;
+	int i, j, n, pc, plen, metachar, quote;
 
 	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
 	if (len < 0) len = strlen(s);
@@ -1084,23 +1109,15 @@ int len;
 	for (i = 0; i < len; i++) {
 		cp = NULL;
 		metachar = 0;
-		if (s[i] == quote) {
-			quote = '\0';
-			continue;
-		}
-		else if (quote == '\'');
-		else if (isnmeta(s, i, quote, len)) {
+		pc = parsechar(&(s[i]), len - i, '\0', 0, &quote, NULL);
+		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) continue;
+		else if (pc == PC_META) {
 			metachar = 1;
 			i++;
 		}
-		else if (quote);
-		else if (s[i] == '\'' || s[i] == '"') {
-			quote = s[i];
-			continue;
-		}
 
 		if (paren) {
-			paren = c_realloc(paren, plen + 1, size);
+			paren = c_realloc(paren, plen + 1, &size);
 
 			if (quote || metachar) {
 #ifdef	BASHSTYLE
@@ -1195,7 +1212,7 @@ int len;
 				cp = wildsymbol2;
 				break;
 			case '[':
-				paren = c_malloc(size);
+				paren = c_realloc(NULL, 0, &size);
 				plen = 0;
 				break;
 			default:
@@ -1223,8 +1240,8 @@ int len;
 			else cp[j++] = toupper2(s[i]);
 			cp[j] = '\0';
 		}
-		re[n] = cp;
-		re = b_realloc(re, ++n, reg_t);
+		re[n++] = cp;
+		re = b_realloc(re, n, reg_t);
 		re[n] = NULL;
 	}
 	if (paren) {
@@ -1234,8 +1251,8 @@ int len;
 			return(NULL);
 		}
 		cp = strdup2("[");
-		re[n] = cp;
-		re = b_realloc(re, ++n, reg_t);
+		re[n++] = cp;
+		re = b_realloc(re, n, reg_t);
 		re[n] = NULL;
 	}
 	return((reg_t *)realloc2(re, (n + 1) * sizeof(reg_t)));
@@ -1772,12 +1789,14 @@ char *com, *search;
 #ifndef	_NOUSEHASH
 	static char *searchcwd = NULL;
 	char buf[MAXPATHLEN];
-	int n;
+	int n, duperrno;
 
 	if (!com && !search) {
+		duperrno = errno;
 		freehash(hpp);
 		if (searchcwd) free(searchcwd);
 		searchcwd = NULL;
+		errno = duperrno;
 		return(CM_NOTFOUND);
 	}
 #endif
@@ -1805,7 +1824,7 @@ char *com, *search;
 	if ((*hpp = findhash(com, n))) {
 		path = (*hpp) -> path;
 		if (((*hpp) -> type & CM_RECALC)
-		&& (!searchcwd || strcmp(searchcwd, Xgetwd(buf))));
+		&& (!searchcwd || !Xgetwd(buf) || strcmp(searchcwd, buf)));
 		else if (isexecute(path, 0, 1) >= 0) return((*hpp) -> type);
 		rmhash(com, n);
 	}
@@ -2182,12 +2201,51 @@ char **var;
 VOID freevar(var)
 char **var;
 {
-	int i;
+	int i, duperrno;
 
+	duperrno = errno;
 	if (var) {
 		for (i = 0; var[i]; i++) free(var[i]);
 		free(var);
 	}
+	errno = duperrno;
+}
+
+int parsechar(s, len, spc, qflag, qp, pqp)
+char *s;
+int len, spc, qflag, *qp, *pqp;
+{
+	if (*s == *qp) {
+		if (!pqp) *qp = '\0';
+		else {
+			*qp = *pqp;
+			*pqp = '\0';
+		}
+		return(PC_CLQUOTE);
+	}
+#ifdef	CODEEUC
+	if (isekana(s, 0)) return(PC_WORD);
+#endif
+	if (iskanji1(s, 0)) return(PC_WORD);
+	if (*qp == '`') return(PC_BQUOTE);
+	if (*qp == '\'') return(PC_SQUOTE);
+	if (spc && *s == spc) return(*s);
+	if (ismeta(s, 0, *qp, len)) return(PC_META);
+	if (qflag > 0 && *s == '`') {
+		if (pqp && *qp) *pqp = *qp;
+		*qp = *s;
+		return(PC_OPQUOTE);
+	}
+	if (*qp) return(PC_DQUOTE);
+	if (qflag >= 0 && *s == '\'') {
+		*qp = *s;
+		return(PC_OPQUOTE);
+	}
+	if (*s == '"') {
+		*qp = *s;
+		return(PC_OPQUOTE);
+	}
+	return(PC_NORMAL);
 }
 
 static int NEAR skipvar(bufp, eolp, ptrp, qed)
@@ -2239,7 +2297,8 @@ int *eolp, *ptrp, qed;
 #endif
 	if (isidentchar((*bufp)[*ptrp])) {
 		while ((*bufp)[++(*ptrp)])
-			if (!isidentchar((*bufp)[*ptrp])) break;
+			if (!isidentchar((*bufp)[*ptrp])
+			&& !isdigit((*bufp)[*ptrp])) break;
 	}
 	else (*ptrp)++;
 
@@ -2299,38 +2358,26 @@ char *next;
 int qed, nonl, nest;
 #endif
 {
-	char *cp;
-	int q, len;
 #ifdef	BASHSTYLE
 	int pq;
 #endif
+	char *cp;
+	int pc, q, len;
 
 #ifdef	BASHSTYLE
-	pq =
+	pq = '\0';
 #endif
 	q = '\0';
 	len = strlen(next);
 	while (s[*ptrp]) {
-		if (s[*ptrp] == q) {
 #ifdef	BASHSTYLE
 	/* bash can include `...` in "..." */
-			q = pq;
-			pq = '\0';
+		pc = parsechar(&(s[*ptrp]), -1, '$', 1, &q, &pq);
 #else
-			q = '\0';
+		pc = parsechar(&(s[*ptrp]), -1, '$', 1, &q, NULL);
 #endif
-		}
-		else if (q == '`');
-#ifdef	CODEEUC
-		else if (isekana(s, *ptrp)) {
-			if (s[*ptrp + 1]) (*ptrp)++;
-		}
-#endif
-		else if (iskanji1(s, *ptrp)) {
-			if (s[*ptrp + 1]) (*ptrp)++;
-		}
-		else if (q == '\'');
-		else if (s[*ptrp] == '$') {
+		if (pc == PC_WORD || pc == PC_META) (*ptrp)++;
+		else if (pc == '$') {
 			if (!s[++(*ptrp)]) return(0);
 			cp = s;
 			if (skipvar(&cp, NULL, ptrp, (q) ? q : qed) < 0)
@@ -2338,22 +2385,15 @@ int qed, nonl, nest;
 			*ptrp += (cp - s);
 			continue;
 		}
-		else if (ismeta(s, *ptrp, q)) (*ptrp)++;
-		else if (s[*ptrp] == '`') {
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
-			if (q) pq = q;
-#endif
-			q = s[*ptrp];
-		}
-		else if (q);
-		else if (s[*ptrp] == '\'' || s[*ptrp] == '"') q = s[*ptrp];
+		else if (pc != PC_NORMAL);
 		else if (nonl && s[*ptrp] == '\n') return(-1);
 #ifndef	MINIMUMSHELL
 		else if (nest && s[*ptrp] == '(') {
+			int tmpq;
+
+			tmpq = (q) ? q : qed;
 			(*ptrp)++;
-			if (skipvarvalue(s, ptrp, ")",
-			(q) ? q : qed, nonl, nest) < 0)
+			if (skipvarvalue(s, ptrp, ")", tmpq, nonl, nest) < 0)
 				return(-1);
 			continue;
 		}
@@ -2375,39 +2415,43 @@ int plen, mode;
 {
 	reg_t *re;
 	char *cp, *ret, *tmp;
-	int c;
+	int c, len;
 
-	if (!s || !*s || !(re = regexp_init(pattern, plen))) return(NULL);
+	if (!s || !*s) return(NULL);
+	tmp = strdupcpy(pattern, plen);
+	pattern = evalarg(tmp, 0, 1, '\0');
+	free(tmp);
+	re = regexp_init(pattern, -1);
+	free(pattern);
+	if (!re) return(NULL);
 	ret = NULL;
+	len = strlen(s);
 	if ((mode & ~0x80) != '#') {
-		if (mode & 0x80) for (cp = s; *cp; cp++) {
+		if (mode & 0x80) for (cp = s; cp < s + len; cp++) {
 			if (regexp_exec(re, cp, 0)) {
 				ret = cp;
 				break;
 			}
 		}
-		else for (cp = s + strlen(s) - 1; cp >= s; cp--) {
+		else for (cp = s + len - 1; cp >= s; cp--) {
 			if (regexp_exec(re, cp, 0)) {
 				ret = cp;
 				break;
 			}
 		}
-		if (ret) {
-			ret = malloc2(cp - s + 1);
-			strncpy2(ret, s, cp - s);
-		}
+		if (ret) ret = strdupcpy(s, ret - s);
 	}
 	else {
 		tmp = strdup2(s);
 		if (mode & 0x80)
-		for (cp = tmp + strlen(tmp); cp >= tmp; cp--) {
+		for (cp = tmp + len; cp >= tmp; cp--) {
 			*cp = '\0';
 			if (regexp_exec(re, tmp, 0)) {
 				ret = cp;
 				break;
 			}
 		}
-		else for (cp = tmp + 1; *cp; cp++) {
+		else for (cp = tmp + 1; cp <= tmp + len; cp++) {
 			c = *cp;
 			*cp = '\0';
 			if (regexp_exec(re, tmp, 0)) {
@@ -2416,7 +2460,7 @@ int plen, mode;
 			}
 			*cp = c;
 		}
-		if (ret) ret = strdup2(&(s[cp - tmp]));
+		if (ret) ret = strdup2(&(s[ret - tmp]));
 		free(tmp);
 	}
 	regexp_free(re);
@@ -2574,9 +2618,12 @@ int s, len, vlen, mode, quoted;
 	val = strdupcpy(&(arg[s]), vlen);
 	*cpp = evalarg(val, (mode == '=' || mode == '?'), 1, quoted);
 	free(val);
-	if (!(*cpp)) return(-1);
+	if (!*cpp) return(-1);
 
 	if (mode == '=') {
+#ifdef	FD
+		demacroarg(cpp);
+#endif
 		if (setvar(arg, *cpp, len) < 0) {
 			free(*cpp);
 			*cpp = NULL;
@@ -2588,10 +2635,16 @@ int s, len, vlen, mode, quoted;
 		val = strdupcpy(&(arg[s]), vlen);
 		*cpp = evalarg(val, 0, 1, quoted);
 		free(val);
-		if (!(*cpp)) return(-1);
+		if (!*cpp) return(-1);
+# ifdef	FD
+		demacroarg(cpp);
+# endif
 #endif
 	}
 	else if (mode == '?') {
+#ifdef	FD
+		demacroarg(cpp);
+#endif
 		for (i = 0; i < len; i++) fputc(arg[i], stderr);
 		fputs(": ", stderr);
 		if (vlen <= 0)
@@ -2661,7 +2714,7 @@ int quoted;
 			if ((new = (*posixsubstfunc)(top, &n))) break;
 			if (n < 0 || !top[n]) return(-1);
 			if (skipvarvalue(top, &n, ")", quoted, 0, 1) < 0)
-                                return(-1);
+				return(-1);
 			*argp = top + n - 1;
 		}
 		mode = '\0';
@@ -2734,8 +2787,8 @@ int quoted;
 		*bufp = insertarg(*bufp, ptr, arg, *argp - arg + 1, vlen);
 		addmeta(&((*bufp)[ptr]), cp, 0, quoted);
 	}
-	else if (!cp);
-	else if (c == '@' && !(*cp) && quoted
+	else if (!cp) vlen = 0;
+	else if (c == '@' && !*cp && quoted
 	&& ptr > 0 && (*bufp)[ptr - 1] == quoted && *(*argp + 1) == quoted) {
 		vlen = 0;
 		ptr--;
@@ -2952,16 +3005,16 @@ char *evalarg(arg, stripq, backq, qed)
 char *arg;
 int stripq, backq, qed;
 {
-#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char path[MAXPATHLEN], alias[MAXPATHLEN];
 #endif
-	char *cp, *buf, *bbuf;
-	int i, j, prev, q;
 #ifdef	BASHSTYLE
 	int pq;
 #endif
+	char *cp, *buf, *bbuf;
+	int i, j, pc, prev, q;
 
-#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	if (*arg == '"' && (i = strlen(arg)) > 2 && arg[i - 1] == '"') {
 		strncpy2(path, &(arg[1]), i - 2);
 		if (shortname(path, alias) == alias) {
@@ -2982,30 +3035,28 @@ int stripq, backq, qed;
 	if (!backquotefunc) backq = 0;
 	bbuf = (backq) ? malloc2(i) : NULL;
 #ifdef	BASHSTYLE
-	pq =
+	pq = '\0';
 #endif
 	prev = q = '\0';
 	i = j = 0;
 
 	for (cp = arg; *cp; prev = *(cp++)) {
-		if (*cp == q) {
-			if (q == '`') {
+#ifdef	BASHSTYLE
+	/* bash can include `...` in "..." */
+		pc = parsechar(cp, -1, '$', backq, &q, &pq);
+#else
+		pc = parsechar(cp, -1, '$', backq, &q, NULL);
+#endif
+		if (pc == PC_CLQUOTE) {
+			if (*cp == '`') {
 				bbuf[j] = '\0';
 				buf = replacebackquote(buf, &i,
 					bbuf, strlen(cp + 1));
 				j = 0;
 			}
 			else if (!stripq) buf[i++] = *cp;
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
-			q = pq;
-			pq = '\0';
-#else
-			q = '\0';
-#endif
 		}
-#ifdef	CODEEUC
-		else if (isekana(cp, 0)) {
+		else if (pc == PC_WORD) {
 			if (q == '`') {
 				bbuf[j++] = *cp++;
 				bbuf[j++] = *cp;
@@ -3015,32 +3066,24 @@ int stripq, backq, qed;
 				buf[i++] = *cp;
 			}
 		}
-#endif
-		else if (iskanji1(cp, 0)) {
-			if (q == '`') {
-				bbuf[j++] = *cp++;
-				bbuf[j++] = *cp;
-			}
-			else {
-				buf[i++] = *cp++;
-				buf[i++] = *cp;
-			}
-		}
-		else if (q == '`') bbuf[j++] = *cp;
-		else if (q == '\'') buf[i++] = *cp;
-		else if (*cp == '$' && cp[1]) {
+		else if (pc == PC_BQUOTE) bbuf[j++] = *cp;
+		else if (pc == PC_SQUOTE || pc == PC_DQUOTE) buf[i++] = *cp;
+		else if (pc == '$') {
+			int tmpq;
+
+			tmpq = (q) ? q : qed;
+			if (!cp[1]) buf[i++] = *cp;
 #ifdef	FAKEMETA
-			if (cp[1] == '$' || cp[1] == '~') cp++;
-			else
+			else if (cp[1] == '$' || cp[1] == '~') cp++;
 #endif
-			if ((i = evalvar(&buf, i, &cp, (q) ? q : qed)) < 0) {
+			else if ((i = evalvar(&buf, i, &cp, tmpq)) < 0) {
 				if (bbuf) free(bbuf);
 				free(buf);
 				if (i < -1) *arg = '\0';
 				return(NULL);
 			}
 		}
-		else if (ismeta(cp, 0, q)) {
+		else if (pc == PC_META) {
 			cp++;
 			if (*cp == '$');
 			else if (backq && *cp == '`');
@@ -3048,19 +3091,11 @@ int stripq, backq, qed;
 			else buf[i++] = PMETA;
 			buf[i++] = *cp;
 		}
-		else if (backq && *cp == '`') {
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
-			if (q) pq = q;
-#endif
-			q = *cp;
-			j = 0;
-		}
-		else if (q) buf[i++] = *cp;
-		else if (*cp == '\'' || *cp == '"') {
-			q = *cp;
-			if (!stripq) buf[i++] = *cp;
+		else if (pc == PC_OPQUOTE) {
+			if (*cp == '`') j = 0;
+			else if (!stripq) buf[i++] = *cp;
 		}
+		else if (pc != PC_NORMAL);
 		else if (*cp == '~' && (!prev || prev == ':' || prev == '='))
 			i = evalhome(&buf, i, &cp);
 		else buf[i++] = *cp;
@@ -3082,21 +3117,14 @@ int argc;
 char ***argvp, *ifs;
 {
 	char *cp;
-	int i, j, n, quote;
+	int i, j, n, pc, quote;
 
 	for (n = 0; n < argc; n++) {
 		for (i = 0, quote = '\0'; (*argvp)[n][i]; i++) {
-			if ((*argvp)[n][i] == quote) quote = '\0';
-#ifdef	CODEEUC
-			else if (isekana((*argvp)[n], i)) i++;
-#endif
-			else if (iskanji1((*argvp)[n], i)) i++;
-			else if (quote == '\'');
-			else if (ismeta((*argvp)[n], i, quote)) i++;
-			else if (quote);
-			else if ((*argvp)[n][i] == '\''
-			|| (*argvp)[n][i] == '"')
-				quote = (*argvp)[n][i];
+			pc = parsechar(&((*argvp)[n][i]), -1,
+				'\0', 0, &quote, NULL);
+			if (pc == PC_WORD || pc == PC_META) i++;
+			else if (pc != PC_NORMAL);
 			else if (strchr(ifs, (*argvp)[n][i])) {
 				for (j = i + 1; (*argvp)[n][j]; j++)
 					if (!strchr(ifs, (*argvp)[n][j]))
@@ -3137,35 +3165,26 @@ char ***argvp;
 int stripq;
 {
 	char *cp, **wild;
-	int i, j, n, w, size, quote;
+	ALLOC_T size;
+	int i, j, n, pc, w, quote;
 
 	for (n = 0; n < argc; n++) {
-		cp = c_malloc(size);
+		cp = c_realloc(NULL, 0, &size);
 		for (i = j = w = 0, quote = '\0'; (*argvp)[n][i]; i++) {
-			cp = c_realloc(cp, j + 1, size);
-			if ((*argvp)[n][i] == quote) {
-				quote = '\0';
+			cp = c_realloc(cp, j + 1, &size);
+			pc = parsechar(&((*argvp)[n][i]), -1,
+				'\0', 0, &quote, NULL);
+			if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) {
 				if (stripq) continue;
 			}
-#ifdef	CODEEUC
-			else if (isekana((*argvp)[n], i))
-				cp[j++] = (*argvp)[n][i++];
-#endif
-			else if (iskanji1((*argvp)[n], i))
-				cp[j++] = (*argvp)[n][i++];
-			else if (quote == '\'');
-			else if (ismeta((*argvp)[n], i, quote)) {
+			else if (pc == PC_WORD) cp[j++] = (*argvp)[n][i++];
+			else if (pc == PC_META) {
 				i++;
 				if (quote
 				&& !strchr(DQ_METACHAR, (*argvp)[n][i]))
 					cp[j++] = PMETA;
 			}
-			else if (quote);
-			else if ((*argvp)[n][i] == '\''
-			|| (*argvp)[n][i] == '"') {
-				quote = (*argvp)[n][i];
-				if (stripq) continue;
-			}
+			else if (pc != PC_NORMAL);
 			else if (!strchr("?*[", (*argvp)[n][i]));
 			else if ((wild = evalwild((*argvp)[n]))) {
 				w = countvar(wild);
@@ -3207,33 +3226,23 @@ int stripquote(arg, stripq)
 char *arg;
 int stripq;
 {
-	int i, j, quote, stripped;
+	int i, j, pc, quote, stripped;
 
 	stripped = 0;
 	if (!arg) return(stripped);
 	for (i = j = 0, quote = '\0'; arg[i]; i++) {
-		if (arg[i] == quote) {
-			quote = '\0';
+		pc = parsechar(&(arg[i]), -1, '\0', 0, &quote, NULL);
+		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) {
 			stripped = 1;
 			if (stripq) continue;
 		}
-#ifdef	CODEEUC
-		else if (isekana(arg, i)) arg[j++] = arg[i++];
-#endif
-		else if (iskanji1(arg, i)) arg[j++] = arg[i++];
-		else if (quote == '\'');
-		else if (ismeta(arg, i, quote)) {
+		else if (pc == PC_WORD) arg[j++] = arg[i++];
+		else if (pc == PC_META) {
 			i++;
 			stripped = 1;
 			if (quote && arg[i] != quote && arg[i] != PMETA)
 				arg[j++] = PMETA;
 		}
-		else if (quote);
-		else if (arg[i] == '\'' || arg[i] == '"') {
-			quote = arg[i];
-			stripped = 1;
-			if (stripq) continue;
-		}
 
 		arg[j++] = arg[i];
 	}
@@ -3245,12 +3254,12 @@ char *_evalpath(path, eol, uniqdelim, evalq)
 char *path, *eol;
 int uniqdelim, evalq;
 {
-#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char alias[MAXPATHLEN];
 	int top = -1;
 #endif
 	char *cp, *tmp;
-	int i, j, c, size, quote;
+	int i, j, c, pc, size, quote;
 
 	if (eol) i = eol - path;
 	else i = strlen(path);
@@ -3267,8 +3276,9 @@ int uniqdelim, evalq;
 	tmp = malloc2(size);
 	quote = '\0';
 	for (i = j = c = 0; cp[i]; c = cp[i++]) {
-		if (cp[i] == quote) {
-#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+		pc = parsechar(&(cp[i]), -1, '\0', 0, &quote, NULL);
+		if (pc == PC_CLQUOTE) {
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 			if (!evalq && top >= 0 && ++top < j) {
 				tmp[j] = '\0';
 				if (shortname(&(tmp[top]), alias) == alias) {
@@ -3283,28 +3293,22 @@ int uniqdelim, evalq;
 			}
 			top = -1;
 #endif
-			quote = '\0';
 			if (evalq) continue;
 		}
-#ifdef	CODEEUC
-		else if (isekana(cp, i)) tmp[j++] = cp[i++];
-#endif
-		else if (iskanji1(cp, i)) tmp[j++] = cp[i++];
-		else if (quote == '\'');
-		else if (ismeta(cp, i, quote)) {
+		else if (pc == PC_WORD) tmp[j++] = cp[i++];
+		else if (pc == PC_META) {
 			i++;
 			if (!evalq
 			|| (quote && !strchr(DQ_METACHAR, cp[i])))
 				tmp[j++] = PMETA;
 		}
-		else if (quote);
-		else if (cp[i] == '\'' || cp[i] == '"') {
-#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+		else if (pc == PC_OPQUOTE) {
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 			if (cp[i] == '"') top = j;
 #endif
-			quote = cp[i];
 			if (evalq) continue;
 		}
+		else if (pc != PC_NORMAL);
 		else if (uniqdelim && cp[i] == _SC_ && c == _SC_) continue;
 		tmp[j++] = cp[i];
 	}
@@ -3319,7 +3323,7 @@ int uniqdelim;
 {
 	char *cp;
 
-	if (!path || !(*path)) return(path);
+	if (!path || !*path) return(path);
 	for (cp = path; *cp == ' ' || *cp == '\t'; cp++);
 	cp = _evalpath(cp, NULL, uniqdelim, 1);
 	free(path);
