@@ -13,7 +13,7 @@
 #define	PATHDELIM	':'
 #define	PMETA		META
 #define	METACHAR	"\t\n !\"#$&'()*;<=>?[\\]`|"
-#define	DQ_METACHAR	"\"$`"
+#define	DQ_METACHAR	"\"$\\`"
 #else	/* MSDOS */
 # ifdef	_NOORIGSHELL
 # define	FAKEMETA
@@ -24,11 +24,20 @@
 # else
 # define	PMETA		'%'
 # define	METACHAR	"\t\n !\"#$%&'()*;<=>?[]`|"
-# define	DQ_METACHAR	"\"$`"
+# define	DQ_METACHAR	"\"$%`"
 # define	PATHDELIM	':'
 # endif
 #endif	/* MSDOS */
 
+#define	PC_NORMAL	0
+#define	PC_OPQUOTE	1
+#define	PC_CLQUOTE	2
+#define	PC_SQUOTE	3
+#define	PC_DQUOTE	4
+#define	PC_BQUOTE	5
+#define	PC_WORD		6
+#define	PC_META		7
+
 #define	MAXLONGWIDTH	20		/* log10(2^64) = 19.266 */
 #ifdef	LSI_C
 #define	toupper2	toupper
@@ -89,8 +98,8 @@ typedef struct _hashlist {
 
 #if	!MSDOS
 typedef struct _devino_t {
-	dev_t	dev;
-	ino_t	ino;
+	dev_t dev;
+	ino_t ino;
 } devino_t;
 #endif
 
@@ -99,6 +108,13 @@ typedef u_short	uid_t;
 typedef u_short	gid_t;
 #endif
 
+#ifdef	USEPID_T
+typedef	pid_t		p_id_t;
+# else
+typedef	long		p_id_t;
+#endif
+
+#ifndef	VOID_P
 #ifdef	NOVOID
 #define	VOID
 #define	VOID_T	int
@@ -108,6 +124,25 @@ typedef u_short	gid_t;
 #define	VOID_T	void
 #define	VOID_P	void *
 #endif
+#endif
+
+#if	defined (SIGARGINT) || defined (NOVOID)
+#define	sigarg_t	int
+#else
+#define	sigarg_t	void
+#endif
+
+#ifdef	SIGFNCINT
+#define	sigfnc_t	int
+#else
+# ifdef	NOVOID
+# define	sigfnc_t
+# else
+# define	sigfnc_t	void
+# endif
+#endif
+
+typedef sigarg_t (*sigcst_t)__P_((sigfnc_t));
 
 typedef struct _uidtable {
 	uid_t uid;
@@ -134,19 +169,17 @@ extern char *strcatdelim2 __P_((char *, char *, char *));
 #if	MSDOS
 #define	strnpathcmp	strnpathcmp2
 #define	strpathcmp	strpathcmp2
-#define	strcasecmp2	stricmp
 #else
 #define	strnpathcmp	strncmp
 #define	strpathcmp	strcmp
-extern int strcasecmp2 __P_((char *, char *));
 #endif
+extern int strcasecmp2 __P_((char *, char *));
 extern int strpathcmp2 __P_((char *, char *));
 extern int strnpathcmp2 __P_((char *, char *, int));
 extern char *strdupcpy __P_((char *, int));
 extern int isidentchar __P_((int));
 extern int isdotdir __P_((char *));
-extern int ismeta __P_((char *s, int, int));
-extern int isnmeta __P_((char *s, int, int, int));
+extern char *isrootdir __P_((char *));
 extern reg_t *regexp_init __P_((char *, int));
 extern int regexp_exec __P_((reg_t *, char *, int));
 extern VOID regexp_free __P_((reg_t *));
@@ -165,6 +198,7 @@ extern char *findcommon __P_((int, char **));
 extern char *catvar __P_((char *[], int));
 extern int countvar __P_((char **));
 extern VOID freevar __P_((char **));
+extern int parsechar __P_((char *, int, int, int, int *, int *));
 #if	!MSDOS
 extern uidtable *finduid __P_((uid_t, char *));
 extern gidtable *findgid __P_((gid_t, char *));
@@ -188,8 +222,8 @@ extern hashlist **hashtable;
 extern char *(*getvarfunc)__P_((char *, int));
 extern int (*putvarfunc)__P_((char *, int));
 extern int (*getretvalfunc)__P_((VOID_A));
-extern long (*getpidfunc)__P_((VOID_A));
-extern long (*getlastpidfunc)__P_((VOID_A));
+extern p_id_t (*getpidfunc)__P_((VOID_A));
+extern p_id_t (*getlastpidfunc)__P_((VOID_A));
 extern char *(*getflagfunc)__P_((VOID_A));
 extern int (*checkundeffunc)__P_((char *, char *, int));
 extern VOID (*exitfunc)__P_((VOID_A));
