@@ -5,7 +5,6 @@
  */
 
 #include <ctype.h>
-#include <signal.h>
 #include "fd.h"
 #include "term.h"
 #include "func.h"
@@ -314,8 +313,10 @@ static VOID NEAR showtree(VOID_A)
 		locate(0, i);
 		putterm(l_clear);
 	}
-	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++)
-		sprintf(bufptr(i), "%-*.*s", TREEFIELD, TREEFIELD, " ");
+	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++) {
+		memset(bufptr(i), ' ', TREEFIELD);
+		bufptr(i)[TREEFIELD] = '\0';
+	}
 	_showtree(tr_root -> sub, 1, 0, tr_top);
 	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++) {
 		locate(1, i);
@@ -332,13 +333,13 @@ static VOID NEAR treebar(VOID_A)
 	locate(1, LFILETOP);
 	cputs2("Tree=");
 	kanjiputs2(treepath, n_column - 6, 0);
-	locate(0, LMESLINE);
+	locate(0, L_MESLINE);
 	putterm(l_clear);
 }
 
 VOID rewritetree(VOID_A)
 {
-	tr_scr = realloc2(tr_scr, (FILEPERLOW - 1) * (TREEFIELD + 1));
+	tr_scr = realloc2(tr_scr, (FILEPERROW - 1) * (TREEFIELD + 1));
 	searchtree();
 	showtree();
 	treebar();
@@ -551,13 +552,9 @@ static int NEAR _tree_input(VOID_A)
 	int ch, tmp, half;
 
 	keyflush();
-#ifdef	_NOEDITMODE
-	ch = Xgetkey(SIGALRM, 0);
-#else
 	Xgetkey(-1, 0);
-	ch = Xgetkey(SIGALRM, 0);
+	ch = Xgetkey(1, 0);
 	Xgetkey(-1, 0);
-#endif
 
 	old = tr_cur;
 	switch (ch) {
@@ -568,7 +565,7 @@ static int NEAR _tree_input(VOID_A)
 			treedown();
 			break;
 		case K_PPAGE:
-			half = (FILEPERLOW - 1) / 2;
+			half = (FILEPERROW - 1) / 2;
 			tmp = LFILETOP + half + 1 - tr_line;
 			if (tr_top + half > LFILETOP + 1)
 				half = LFILETOP - tr_top + 1;
@@ -578,7 +575,7 @@ static int NEAR _tree_input(VOID_A)
 			searchtree();
 			break;
 		case K_NPAGE:
-			half = (FILEPERLOW - 1) / 2;
+			half = (FILEPERROW - 1) / 2;
 			tmp = half + (LFILETOP + half + 1) - tr_line;
 			while (tmp-- > 0) if (treedown() < 0) break;
 			tmp = tr_line - (LFILETOP + half + 1);
@@ -662,7 +659,7 @@ static int NEAR _tree_input(VOID_A)
 			ch = '\0';
 			free(cwd);
 			break;
-		case CTRL('L'):
+		case K_CTRL('L'):
 			rewritefile(1);
 			break;
 		case K_ESC:
@@ -751,7 +748,7 @@ static char *NEAR _tree(VOID_A)
 		tr_line = LFILEBOTTOM - 2;
 	}
 
-	tr_scr = malloc2((FILEPERLOW - 1) * (TREEFIELD + 1));
+	tr_scr = malloc2((FILEPERROW - 1) * (TREEFIELD + 1));
 	searchtree();
 	showtree();
 	win_x = 0;
@@ -796,6 +793,10 @@ int cleanup, *ddp;
 {
 	char *path, *dupfullpath;
 
+	if (FILEPERROW < WFILEMINTREE) {
+		warning(0, NOROW_K);
+		return(NULL);
+	}
 	dupfullpath = strdup2(fullpath);
 	do {
 		path = _tree();
