@@ -80,8 +80,7 @@
 #endif
 #define	MINTYPE(t)	((t)(-1L << (BITSPERBYTE * sizeof(t) - 1)))
 #define	MAXTYPE(t)	((t)~MINTYPE(t))
-#define	MAXNUMCOLS	10
-#define	MAXCOLSCOMMA(d)	(MAXNUMCOLS + (MAXNUMCOLS / (d)))
+#define	MAXCOLSCOMMA(d)	(MAXLONGWIDTH + (MAXLONGWIDTH / (d)))
 #define	MAXHISTNO	MAXTYPE(short)
 
 typedef struct _namelist {
@@ -92,6 +91,9 @@ typedef struct _namelist {
 #if	!MSDOS
 	uid_t st_uid;
 	gid_t st_gid;
+# ifndef	_NOARCHIVE
+	char *linkname;
+# endif
 #endif
 #ifdef	HAVEFLAGS
 	u_long st_flags;
@@ -125,12 +127,6 @@ typedef struct _namelist {
 #define	isarg(file)		((file) -> tmpflags & F_ISARG)
 #define	havestat(file)		((file) -> tmpflags & F_STAT)
 
-typedef struct _assoclist {
-	char *org;
-	char *assoc;
-	struct _assoclist *next;
-} assoclist;
-
 typedef struct _strtable {
 	u_short no;
 	char *str;
@@ -154,7 +150,7 @@ typedef struct _functable {
 #ifndef	_NOJPNMES
 	char *hmes;
 #endif
-#ifndef	_NOENGMES
+#if	!defined (_NOENGMES) || defined (_NOJPNMES)
 	char *hmes_eng;
 #endif
 	u_char stat;
@@ -178,7 +174,9 @@ typedef struct _launchtable {
 	char *ext;
 	char *comm;
 # if	FD >= 2
-	char *format;
+	char **format;
+	char **lignore;
+	char **lerror;
 # endif
 	u_char topskip;
 	u_char bottomskip;
@@ -202,6 +200,10 @@ typedef struct _launchtable {
 #define	F_TIME	7
 #define	F_NAME	8
 #define	LF_IGNORECASE	0001
+#define	LF_DIRLOOP	0002
+#define	LF_DIRNOPREP	0004
+#define	LF_FILELOOP	0010
+#define	LF_FILENOPREP	0020
 
 typedef struct _archivetable {
 	char *ext;
@@ -235,10 +237,14 @@ typedef struct _winvartable {
 	launchtable *v_launchp;
 	namelist *v_arcflist;
 	int v_maxarcf;
-# ifndef	_NODOSDRIVE
+# if	(!MSDOS || !defined (_NOUSELFN)) && !defined (_NODOSDRIVE)
 	int v_archdrive;
 # endif
-#endif
+# ifndef	_NOBROWSE
+	launchtable *v_browselist;
+	int v_browselevel;
+# endif
+#endif	/* !_NOARCHIVE */
 #ifndef	_NOTREE
 	char *v_treepath;
 #endif
@@ -266,10 +272,14 @@ extern int win;
 #define	launchp		(winvar[win].v_launchp)
 #define	arcflist	(winvar[win].v_arcflist)
 #define	maxarcf		(winvar[win].v_maxarcf)
-# ifndef	_NODOSDRIVE
-#define	archdrive	(winvar[win].v_archdrive)
+# if	(!MSDOS || !defined (_NOUSELFN)) && !defined (_NODOSDRIVE)
+# define	archdrive	(winvar[win].v_archdrive)
 # endif
-#endif
+# ifndef	_NOBROWSE
+# define	browselist	(winvar[win].v_browselist)
+# define	browselevel	(winvar[win].v_browselevel)
+# endif
+#endif	/* !_NOARCHIVE */
 #ifndef	_NOTREE
 #define	treepath	(winvar[win].v_treepath)
 #endif
@@ -283,6 +293,7 @@ extern int win;
 
 typedef struct _macrostat {
 	short addopt;
+	short needburst;
 	short needmark;
 	u_char flags;
 } macrostat;
@@ -294,6 +305,7 @@ typedef struct _macrostat {
 #define	F_TOSFN		0020
 #define	F_ISARCH	0040
 #define	F_BURST		0100
+#define	F_MARK		0200
 
 #ifdef	_NOORIGSHELL
 typedef struct _aliastable {
