@@ -8,7 +8,6 @@
 #include <fcntl.h>
 #include <io.h>
 #include "machine.h"
-#include "unixdisk.h"
 
 #ifdef	NOVOID
 #define	VOID
@@ -18,6 +17,11 @@
 #define	VOID_T	void
 #endif
 
+#if	MSDOS && defined (_NOUSELFN) && !defined (_NODOSDRIVE)
+#define	_NODOSDRIVE
+#endif
+
+#include "unixdisk.h"
 #include "dosdisk.h"
 
 #ifdef	LSI_C
@@ -1601,6 +1605,9 @@ int n, size, iswrite;
 	}
 
 	ret = 0;
+#ifdef	FAKEUNINIT
+	level = 0;	/* fake for -Wuninitialized */
+#endif
 #ifdef	DJGPP
 	if ((long)size * n > tbsize / 2) {
 		if (checkdrive(drv) > 0) while (n-- > 0) {
